# -*- coding: utf-8 -*-
import warnings
from functools import wraps
from hashlib import sha1, new as hashlib_new
from base64 import urlsafe_b64encode, urlsafe_b64decode
from datetime import datetime, tzinfo, timedelta

from .compat import b, s

try:
    import zlib

    binascii = zlib
except ImportError:
    zlib = None
    import binascii

_BLOCK_SIZE = 1024 * 1024 * 4


def urlsafe_base64_encode(data):
    """urlsafe的base64编码:

    对提供的数据进行urlsafe的base64编码。规格参考：
    https://developer.qiniu.com/kodo/manual/1231/appendix#1

    Args:
        data: 待编码的数据，一般为字符串

    Returns:
        编码后的字符串
    """
    ret = urlsafe_b64encode(b(data))
    return s(ret)


def urlsafe_base64_decode(data):
    """urlsafe的base64解码:

    对提供的urlsafe的base64编码的数据进行解码

    Args:
        data: 待解码的数据，一般为字符串

    Returns:
        解码后的字符串。
    """
    ret = urlsafe_b64decode(s(data))
    return ret


def file_crc32(filePath):
    """计算文件的crc32检验码:

    Args:
        filePath: 待计算校验码的文件路径

    Returns:
        文件内容的crc32校验码。
    """
    crc = 0
    with open(filePath, 'rb') as f:
        for block in _file_iter(f, _BLOCK_SIZE):
            crc = binascii.crc32(block, crc) & 0xFFFFFFFF
    return crc


def io_crc32(io_data):
    result = 0
    for d in io_data:
        result = binascii.crc32(d, result) & 0xFFFFFFFF
    return result


def io_md5(io_data):
    h = hashlib_new('md5')
    for d in io_data:
        h.update(d)
    return h.hexdigest()


def crc32(data):
    """计算输入流的crc32检验码:

    Args:
        data: 待计算校验码的字符流

    Returns:
        输入流的crc32校验码。
    """
    return binascii.crc32(b(data)) & 0xffffffff


def _file_iter(input_stream, size, offset=0):
    """读取输入流:

    Args:
        input_stream: 待读取文件的二进制流
        size:         二进制流的大小

    Raises:
        IOError: 文件流读取失败
    """
    input_stream.seek(offset)
    d = input_stream.read(size)
    while d:
        yield d
        d = input_stream.read(size)
    input_stream.seek(0)


def _sha1(data):
    """单块计算hash:

    Args:
        data: 待计算hash的数据

    Returns:
        输入数据计算的hash值
    """
    h = sha1()
    h.update(data)
    return h.digest()


def etag_stream(input_stream):
    """
    计算输入流的etag

    .. deprecated::
        在 v2 分片上传使用 4MB 以外分片大小时无法正常工作

    Parameters
    ----------
    input_stream: io.IOBase
        支持随机访问的文件型对象

    Returns
    -------
    str

    """
    array = [_sha1(block) for block in _file_iter(input_stream, _BLOCK_SIZE)]
    if len(array) == 0:
        array = [_sha1(b'')]
    if len(array) == 1:
        data = array[0]
        prefix = b'\x16'
    else:
        sha1_str = b('').join(array)
        data = _sha1(sha1_str)
        prefix = b'\x96'
    return urlsafe_base64_encode(prefix + data)


def etag(filePath):
    """
    计算文件的etag:

    .. deprecated::
        在 v2 分片上传使用 4MB 以外分片大小时无法正常工作


    Parameters
    ----------
    filePath: str
        待计算 etag 的文件路径

    Returns
    -------
    str
        输入文件的etag值
    """
    with open(filePath, 'rb') as f:
        return etag_stream(f)


def entry(bucket, key):
    """计算七牛API中的数据格式:

    entry规格参考 https://developer.qiniu.com/kodo/api/1276/data-format

    Args:
        bucket: 待操作的空间名
        key:    待操作的文件名

    Returns:
        符合七牛API规格的数据格式
    """
    if key is None:
        return urlsafe_base64_encode('{0}'.format(bucket))
    else:
        return urlsafe_base64_encode('{0}:{1}'.format(bucket, key))


def decode_entry(e):
    return (s(urlsafe_base64_decode(e)).split(':') + [None] * 2)[:2]


def rfc_from_timestamp(timestamp):
    """将时间戳转换为HTTP RFC格式

    Args:
        timestamp: 整型Unix时间戳（单位秒）
    """
    last_modified_date = datetime.utcfromtimestamp(timestamp)
    last_modified_str = last_modified_date.strftime(
        '%a, %d %b %Y %H:%M:%S GMT')
    return last_modified_str


def _valid_header_key_char(ch):
    is_token_table = [
        "!", "#", "$", "%", "&", "\\", "*", "+", "-", ".",
        "0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
        "A", "B", "C", "D", "E", "F", "G", "H", "I", "J",
        "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T",
        "U", "W", "V", "X", "Y", "Z",
        "^", "_", "`",
        "a", "b", "c", "d", "e", "f", "g", "h", "i", "j",
        "k", "l", "m", "n", "o", "p", "q", "r", "s", "t",
        "u", "v", "w", "x", "y", "z",
        "|", "~"]
    return 0 <= ord(ch) < 128 and ch in is_token_table


def canonical_mime_header_key(field_name):
    for ch in field_name:
        if not _valid_header_key_char(ch):
            return field_name
    result = ""
    upper = True
    for ch in field_name:
        if upper and "a" <= ch <= "z":
            result += ch.upper()
        elif not upper and "A" <= ch <= "Z":
            result += ch.lower()
        else:
            result += ch
        upper = ch == "-"
    return result


class _UTC_TZINFO(tzinfo):
    def utcoffset(self, dt):
        return timedelta(hours=0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


def dt2ts(dt):
    """
    converte datetime to timestamp

    Parameters
    ----------
    dt: datetime.datetime
    """
    if not dt.tzinfo:
        st = (dt - datetime(1970, 1, 1)).total_seconds()
    else:
        st = (dt - datetime(1970, 1, 1, tzinfo=_UTC_TZINFO())).total_seconds()

    return int(st)


def deprecated(reason):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                "DEPRECATED: {} - {}".format(func.__name__, reason),
                DeprecationWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper
    return decorator



import abc
from concurrent import futures

from qiniu.services.storage.uploaders.io_chunked import ChunkInfo
from qiniu.services.storage.uploaders.abc import UploaderBase


class ResumeUploaderBase(UploaderBase):
    """
    Attributes
    ----------
    part_size: int, optional
    progress_handler: function, optional
    upload_progress_recorder: UploadProgressRecorder, optional
    concurrent_executor: futures.Executor, optional
    """
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        bucket_name,
        **kwargs
    ):
        """
        Parameters
        ----------
        bucket_name
        part_size: int
        progress_handler: function
        upload_progress_recorder: UploadProgressRecorder
        max_concurrent_workers: int
        concurrent_executor: futures.Executor
        kwargs
        """
        super(ResumeUploaderBase, self).__init__(bucket_name, **kwargs)

        self.part_size = kwargs.get('part_size', 4 * (1024 ** 2))

        self.progress_handler = kwargs.get(
            'progress_handler',
            None
        )

        self.upload_progress_recorder = kwargs.get(
            'upload_progress_recorder',
            None
        )

        max_workers = kwargs.get('max_concurrent_workers', 3)
        self.concurrent_executor = kwargs.get(
            'concurrent_executor',
            futures.ThreadPoolExecutor(max_workers=max_workers)
        )

    def gen_chunk_list(self, size, chunk_size=None, uploaded_chunk_no_list=None):
        """
        Parameters
        ----------
        size: int
        chunk_size: int
        uploaded_chunk_no_list: list[int]

        Yields
        -------
            ChunkInfo
        """
        if not chunk_size:
            chunk_size = self.part_size
        if not uploaded_chunk_no_list:
            uploaded_chunk_no_list = []

        for i, chunk_offset in enumerate(range(0, size, chunk_size)):
            chunk_no = i + 1
            if chunk_no in uploaded_chunk_no_list:
                continue
            yield ChunkInfo(
                chunk_no=chunk_no,
                chunk_offset=chunk_offset,
                chunk_size=min(
                    chunk_size,
                    size - chunk_offset
                )
            )

    @abc.abstractmethod
    def _recover_from_record(
        self,
        file_name,
        key,
        context
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: any

        Returns
        -------
        any
        """

    @abc.abstractmethod
    def _set_to_record(
        self,
        file_name,
        key,
        context
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: any
        """

    @abc.abstractmethod
    def _progress_handler(
        self,
        file_name,
        key,
        context,
        uploaded_size,
        total_size
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: any
        uploaded_size: int
        total_size: int
        """

    @abc.abstractmethod
    def initial_parts(
        self,
        up_token,
        key,
        file_path,
        data,
        data_size,
        modify_time,
        part_size,
        file_name,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        key: str
        file_path: str
        data: IOBase
        data_size: int
        modify_time: int
        part_size: int
        file_name: str
        kwargs: dict

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """

    @abc.abstractmethod
    def upload_parts(
        self,
        up_token,
        data,
        data_size,
        context,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        data: IOBase
        data_size: int
        context: any
        kwargs: dict

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """

    @abc.abstractmethod
    def complete_parts(
        self,
        up_token,
        data_size,
        context,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        data_size: int
        context: any
        kwargs: dict

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """



# -*- coding: utf-8 -*-

"""
pythoncompat
"""

import os
import sys

try:
    import simplejson as json
except (ImportError, SyntaxError):
    # simplejson does not support Python 3.2, it thows a SyntaxError
    # because of u'...' Unicode literals.
    import json  # noqa

# -------
# Platform
# -------

is_windows = sys.platform == 'win32'
is_linux = sys.platform == 'linux'
is_macos = sys.platform == 'darwin'

# -------
# Pythons
# -------

_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)


# ---------
# Specifics
# ---------

if is_py2:
    from urllib import urlencode  # noqa
    from urlparse import urlparse  # noqa
    import StringIO
    StringIO = BytesIO = StringIO.StringIO

    builtin_str = str
    bytes = str
    str = unicode  # noqa
    basestring = basestring  # noqa
    numeric_types = (int, long, float)  # noqa

    def b(data):
        return bytes(data)

    def s(data):
        return bytes(data)

    def u(data):
        return unicode(data, 'unicode_escape')  # noqa

    def is_seekable(data):
        try:
            data.seek(0, os.SEEK_CUR)
            return True
        except (AttributeError, IOError):
            return False

elif is_py3:
    from urllib.parse import urlparse, urlencode  # noqa
    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO

    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    numeric_types = (int, float)

    def b(data):
        if isinstance(data, str):
            return data.encode('utf-8')
        return data

    def s(data):
        if isinstance(data, bytes):
            data = data.decode('utf-8')
        return data

    def u(data):
        return data

    def is_seekable(data):
        return data.seekable()



# -*- coding: utf-8 -*-


class GitRepositoryResource(object):
    def __init__(self, url, mount_path, authorization_token=None,
                 repository_type='github_repository'):
        self.url = url
        self.mount_path = mount_path
        self.authorization_token = authorization_token
        self.repository_type = repository_type

    def to_dict(self):
        data = {
            'type': self.repository_type,
            'url': self.url,
            'mount_path': self.mount_path,
        }
        if self.authorization_token is not None:
            data['authorization_token'] = self.authorization_token
        return data


class KodoResource(object):
    def __init__(self, bucket, mount_path, prefix=None, read_only=None):
        self.bucket = bucket
        self.mount_path = mount_path
        self.prefix = prefix
        self.read_only = read_only

    def to_dict(self):
        data = {
            'type': 'kodo',
            'bucket': self.bucket,
            'mount_path': self.mount_path,
        }
        if self.prefix is not None:
            data['prefix'] = self.prefix
        if self.read_only is not None:
            data['read_only'] = self.read_only
        return data



# -*- coding: utf-8 -*-
from qiniu.compat import is_py2, is_py3


class ResponseInfo(object):
    """七牛HTTP请求返回信息类

    该类主要是用于获取和解析对七牛发起各种请求后的响应包的header和body。

    Attributes:
        status_code (int): 整数变量，响应状态码
        text_body (str):   字符串变量，响应的body
        req_id (str):      字符串变量，七牛HTTP扩展字段，参考 https://developer.qiniu.com/kodo/3924/common-request-headers
        x_log (str):       字符串变量，七牛HTTP扩展字段，参考 https://developer.qiniu.com/kodo/3924/common-request-headers
        error (str):       字符串变量，响应的错误内容
    """

    def __init__(self, response, exception=None):
        """用响应包和异常信息初始化ResponseInfo类"""
        self.__response = response
        self.exception = exception
        if response is None:
            self.url = None
            self.status_code = -1
            self.text_body = None
            self.req_id = None
            self.x_log = None
            self.error = str(exception)
        else:
            self.url = response.url
            self.status_code = response.status_code
            self.text_body = response.text
            self.req_id = response.headers.get('X-Reqid')
            self.x_log = response.headers.get('X-Log')
            if self.status_code >= 400:
                if self.__check_json(response):
                    ret = response.json() if response.text != '' else None
                    if ret is None:
                        self.error = 'unknown'
                    else:
                        self.error = response.text
                else:
                    self.error = response.text
            if self.req_id is None and self.status_code == 200:
                self.error = 'server is not qiniu'

    def ok(self):
        return self.status_code // 100 == 2

    def need_retry(self):
        if 100 <= self.status_code < 500:
            return False
        if all([
            self.status_code < 0,
            self.exception is not None,
            'BadStatusLine' in str(self.exception)
        ]):
            return False
        # https://developer.qiniu.com/fusion/kb/1352/the-http-request-return-a-status-code
        # https://developer.qiniu.com/kodo/3928/error-responses
        if self.status_code in [
            501, 509, 573, 579, 608, 612, 614, 616, 618, 630, 631, 632, 640, 701
        ]:
            return False
        return True

    def connect_failed(self):
        return self.__response is None or self.req_id is None

    def json(self):
        try:
            self.__response.encoding = "utf-8"
            return self.__response.json()
        except Exception:
            return {}

    def __str__(self):
        if is_py2:
            return ', '.join(
                ['%s:%s' % item for item in self.__dict__.items()]).encode('utf-8')
        elif is_py3:
            return ', '.join(
                ['%s:%s' % item for item in self.__dict__.items()])

    def __repr__(self):
        return self.__str__()

    def __check_json(self, response):
        try:
            response.json()
            return True
        except Exception:
            return False



# -*- coding: utf-8 -*-
import json

from .util import shell_quote


class ReadyCmd(object):
    def __init__(self, cmd):
        self._cmd = cmd

    def get_cmd(self):
        return self._cmd


def wait_for_port(port):
    port = int(port)
    return ReadyCmd(
        "ss -tuln | awk '{{print $5}}' | grep -E '(^|:){0}$'".format(
            port))


def wait_for_url(url, status_code=200):
    status_code = int(status_code)
    return ReadyCmd(
        '[ "$(curl -s -o /dev/null -w "%{{http_code}}" {0})" = "{1}" ]'.format(
            shell_quote(url),
            status_code,
        )
    )


def wait_for_process(process_name):
    return ReadyCmd('pgrep {0} > /dev/null'.format(shell_quote(process_name)))


def wait_for_file(filename):
    return ReadyCmd('[ -f {0} ]'.format(shell_quote(filename)))


def wait_for_timeout(timeout):
    seconds = max(1, int(timeout) // 1000)
    return ReadyCmd('sleep {0}'.format(seconds))


def _ready_cmd_value(command):
    if hasattr(command, 'get_cmd'):
        return command.get_cmd()
    return command


class Template(object):
    def __init__(self):
        self.build_config = {'steps': []}

    def from_image(self, image, credentials=None):
        self.build_config['fromImage'] = image
        self.build_config.pop('fromTemplate', None)
        if credentials:
            self.build_config['fromImageRegistry'] = credentials
        return self

    fromImage = from_image

    def from_template(self, template_id):
        self.build_config['fromTemplate'] = template_id
        self.build_config.pop('fromImage', None)
        self.build_config.pop('fromImageRegistry', None)
        return self

    fromTemplate = from_template

    def add_step(self, step_type, args, **extra):
        step = {'type': step_type, 'args': [str(arg) for arg in args]}
        step.update(extra)
        self.build_config['steps'].append(step)
        return self

    addStep = add_step

    def run_cmd(self, command, user=None):
        if isinstance(command, (list, tuple)):
            command = ' '.join(shell_quote(arg) for arg in command)
        args = [command]
        if user:
            args.append(user)
        return self.add_step('RUN', args)

    run = run_cmd
    runCmd = run_cmd

    def copy(self, src, dest, chmod=None, chown=None):
        extra = {}
        if chmod is not None:
            extra['chmod'] = str(chmod)
        if chown is not None:
            extra['chown'] = chown
        return self.add_step('COPY', [src, dest], **extra)

    def set_env(self, key, value):
        return self.add_step('ENV', [key, value])

    setEnv = set_env

    def set_start_cmd(self, command, ready_cmd=None):
        self.build_config['startCmd'] = command
        if ready_cmd is not None:
            self.build_config['readyCmd'] = _ready_cmd_value(ready_cmd)
        return self

    setStartCmd = set_start_cmd

    def set_ready_cmd(self, command):
        self.build_config['readyCmd'] = _ready_cmd_value(command)
        return self

    setReadyCmd = set_ready_cmd

    def to_dict(self):
        data = dict(self.build_config)
        data['steps'] = list(self.build_config.get('steps', []))
        return data

    def to_json(self):
        return json.dumps(self.to_dict(), separators=(',', ':'))

    toJSON = to_json



# -*- coding: utf-8 -*-
RS_HOST = 'http://rs.qiniu.com'  # 管理操作Host
RSF_HOST = 'http://rsf.qbox.me'  # 列举操作Host
API_HOST = 'http://api.qiniuapi.com'  # 数据处理操作Host
QUERY_REGION_HOST = 'https://uc.qiniuapi.com'
QUERY_REGION_BACKUP_HOSTS = [
    'kodo-config.qiniuapi.com',
    'uc.qbox.me'
]
UC_HOST = QUERY_REGION_HOST  # 获取空间信息Host
UC_BACKUP_HOSTS = QUERY_REGION_BACKUP_HOSTS

_BLOCK_SIZE = 1024 * 1024 * 4  # 断点续传分块大小，该参数为接口规格，暂不支持修改

_config = {
    'default_zone': None,
    'default_rs_host': RS_HOST,
    'default_rsf_host': RSF_HOST,
    'default_api_host': API_HOST,
    'default_uc_host': UC_HOST,
    'default_uc_backup_hosts': UC_BACKUP_HOSTS,
    'default_query_region_host': QUERY_REGION_HOST,
    'default_query_region_backup_hosts': QUERY_REGION_BACKUP_HOSTS,
    'default_backup_hosts_retry_times': 3,  # 仅控制旧区域 LegacyRegion 查询 Hosts 的重试次数
    'connection_timeout': 30,  # 链接超时为时间为30s
    'connection_retries': 3,  # 链接重试次数为3次
    'connection_pool': 10,  # 链接池个数为10
    'default_upload_threshold': 2 * _BLOCK_SIZE  # put_file上传方式的临界默认值
}

_is_customized_default = {
    k: False
    for k in _config.keys()
}


def is_customized_default(key):
    return _is_customized_default[key]


def get_default(key):
    if key == 'default_zone' and not _is_customized_default[key]:
        # prevent circle import
        from .region import LegacyRegion
        return LegacyRegion()
    return _config[key]


def set_default(
        default_zone=None, connection_retries=None, connection_pool=None,
        connection_timeout=None, default_rs_host=None, default_uc_host=None,
        default_rsf_host=None, default_api_host=None, default_upload_threshold=None,
        default_query_region_host=None, default_query_region_backup_hosts=None,
        default_backup_hosts_retry_times=None, default_uc_backup_hosts=None):
    if default_zone:
        _config['default_zone'] = default_zone
        _is_customized_default['default_zone'] = True
    if default_rs_host:
        _config['default_rs_host'] = default_rs_host
        _is_customized_default['default_rs_host'] = True
    if default_rsf_host:
        _config['default_rsf_host'] = default_rsf_host
        _is_customized_default['default_rsf_host'] = True
    if default_api_host:
        _config['default_api_host'] = default_api_host
        _is_customized_default['default_api_host'] = True
    if default_uc_host:
        _config['default_uc_host'] = default_uc_host
        _is_customized_default['default_uc_host'] = True
        _config['default_uc_backup_hosts'] = []
        _is_customized_default['default_uc_backup_hosts'] = True
        _config['default_query_region_host'] = default_uc_host
        _is_customized_default['default_query_region_host'] = True
        _config['default_query_region_backup_hosts'] = []
        _is_customized_default['default_query_region_backup_hosts'] = True
    if default_uc_backup_hosts is not None:
        _config['default_uc_backup_hosts'] = default_uc_backup_hosts
        _is_customized_default['default_uc_backup_hosts'] = True
        _config['default_query_region_backup_hosts'] = default_uc_backup_hosts
        _is_customized_default['default_query_region_backup_hosts'] = True
    if default_query_region_host:
        _config['default_query_region_host'] = default_query_region_host
        _is_customized_default['default_query_region_host'] = True
        _config['default_query_region_backup_hosts'] = []
        _is_customized_default['default_query_region_backup_hosts'] = True
    if default_query_region_backup_hosts is not None:
        _config['default_query_region_backup_hosts'] = default_query_region_backup_hosts
        _is_customized_default['default_query_region_backup_hosts'] = True
    if default_backup_hosts_retry_times:
        _config['default_backup_hosts_retry_times'] = default_backup_hosts_retry_times
        _is_customized_default['default_backup_hosts_retry_times'] = True
    if connection_retries:
        _config['connection_retries'] = connection_retries
        _is_customized_default['connection_retries'] = True
    if connection_pool:
        _config['connection_pool'] = connection_pool
        _is_customized_default['connection_pool'] = True
    if connection_timeout:
        _config['connection_timeout'] = connection_timeout
        _is_customized_default['connection_timeout'] = True
    if default_upload_threshold:
        _config['default_upload_threshold'] = default_upload_threshold
        _is_customized_default['default_upload_threshold'] = True



# -*- coding: utf-8 -*-
import logging
import platform

import requests
from requests.auth import AuthBase

from qiniu import config, __version__
import qiniu.auth

from .response import ResponseInfo
from .default_client import qn_http_client, _init_http_adapter

_sys_info = '{0}; {1}'.format(platform.system(), platform.machine())
_python_ver = platform.python_version()

USER_AGENT = 'QiniuPython/{0} ({1}; ) Python/{2}'.format(
    __version__, _sys_info, _python_ver)

_session = None
_headers = {'User-Agent': USER_AGENT}


def __return_wrapper(resp):
    if resp.status_code != 200 or resp.headers.get('X-Reqid') is None:
        return None, ResponseInfo(resp)
    resp.encoding = 'utf-8'
    try:
        ret = resp.json()
    except ValueError:
        logging.debug("response body decode error: %s" % resp.text)
        ret = {}
    return ret, ResponseInfo(resp)


def _init():
    global _session
    if _session is None:
        _session = qn_http_client.session
    _init_http_adapter()


def _post(url, data, files, auth, headers=None):
    if _session is None:
        _init()
    try:
        post_headers = _headers.copy()
        if headers is not None:
            for k, v in headers.items():
                post_headers.update({k: v})
        r = _session.post(
            url, data=data, files=files, auth=auth, headers=post_headers,
            timeout=config.get_default('connection_timeout'))
    except Exception as e:
        return None, ResponseInfo(None, e)
    return __return_wrapper(r)


def _put(url, data, files, auth, headers=None):
    if _session is None:
        _init()
    try:
        post_headers = _headers.copy()
        if headers is not None:
            for k, v in headers.items():
                post_headers.update({k: v})
        r = _session.put(
            url, data=data, files=files, auth=auth, headers=post_headers,
            timeout=config.get_default('connection_timeout'))
    except Exception as e:
        return None, ResponseInfo(None, e)
    return __return_wrapper(r)


def _get(url, params, auth, headers=None):
    if _session is None:
        _init()
    try:
        get_headers = _headers.copy()
        if headers is not None:
            for k, v in headers.items():
                get_headers.update({k: v})
        r = _session.get(
            url,
            params=params,
            auth=auth,
            timeout=config.get_default('connection_timeout'),
            headers=get_headers)
    except Exception as e:
        return None, ResponseInfo(None, e)
    return __return_wrapper(r)


class _TokenAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'UpToken {0}'.format(self.token)
        return r


def _post_with_token(url, data, token):
    return _post(url, data, None, _TokenAuth(token))


def _post_with_token_and_headers(url, data, token, headers):
    return _post(url, data, None, _TokenAuth(token), headers)


def _post_file(url, data, files):
    return _post(url, data, files, None)


def _post_with_auth(url, data, auth):
    return _post(url, data, None, qiniu.auth.RequestsAuth(auth))


def _get_with_auth(url, data, auth):
    return _get(url, data, qiniu.auth.RequestsAuth(auth))


def _post_with_auth_and_headers(url, data, auth, headers):
    return _post(url, data, None, qiniu.auth.RequestsAuth(auth), headers)


def _get_with_auth_and_headers(url, data, auth, headers):
    return _get(url, data, qiniu.auth.RequestsAuth(auth), headers)


def _post_with_qiniu_mac_and_headers(url, data, auth, headers):
    return _post(url, data, None, qiniu.auth.QiniuMacRequestsAuth(auth), headers)


def _put_with_auth(url, data, auth):
    return _put(url, data, None, qiniu.auth.RequestsAuth(auth))


def _put_with_token_and_headers(url, data, auth, headers):
    return _put(url, data, None, _TokenAuth(auth), headers)


def _put_with_auth_and_headers(url, data, auth, headers):
    return _put(url, data, None, qiniu.auth.RequestsAuth(auth), headers)


def _put_with_qiniu_mac_and_headers(url, data, auth, headers):
    return _put(url, data, None, qiniu.auth.QiniuMacRequestsAuth(auth), headers)


def _post_with_qiniu_mac(url, data, auth):
    qn_auth = qiniu.auth.QiniuMacRequestsAuth(
        auth
    ) if auth is not None else None

    return _post(url, data, None, qn_auth)


def _get_with_qiniu_mac(url, params, auth):
    qn_auth = qiniu.auth.QiniuMacRequestsAuth(
        auth
    ) if auth is not None else None

    return _get(url, params, qn_auth)


def _get_with_qiniu_mac_and_headers(url, params, auth, headers):
    try:
        post_headers = _headers.copy()
        if headers is not None:
            for k, v in headers.items():
                post_headers.update({k: v})
        r = requests.get(
            url,
            params=params,
            auth=qiniu.auth.QiniuMacRequestsAuth(auth) if auth is not None else None,
            timeout=config.get_default('connection_timeout'),
            headers=post_headers)
    except Exception as e:
        return None, ResponseInfo(None, e)
    return __return_wrapper(r)


def _delete_with_qiniu_mac(url, params, auth):
    try:
        r = requests.delete(
            url,
            params=params,
            auth=qiniu.auth.QiniuMacRequestsAuth(auth) if auth is not None else None,
            timeout=config.get_default('connection_timeout'),
            headers=_headers)
    except Exception as e:
        return None, ResponseInfo(None, e)
    return __return_wrapper(r)


def _delete_with_qiniu_mac_and_headers(url, params, auth, headers):
    try:
        post_headers = _headers.copy()
        if headers is not None:
            for k, v in headers.items():
                post_headers.update({k: v})
        r = requests.delete(
            url,
            params=params,
            auth=qiniu.auth.QiniuMacRequestsAuth(auth) if auth is not None else None,
            timeout=config.get_default('connection_timeout'),
            headers=post_headers)
    except Exception as e:
        return None, ResponseInfo(None, e)
    return __return_wrapper(r)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

from qiniu import etag


def main():
    parser = argparse.ArgumentParser(prog='qiniu')
    sub_parsers = parser.add_subparsers()

    parser_etag = sub_parsers.add_parser(
        'etag',
        description='calculate the etag of the file',
        help='etag [file...]')
    parser_etag.add_argument(
        'etag_files',
        metavar='N',
        nargs='+',
        help='the file list for calculate')

    args = parser.parse_args()

    try:
        etag_files = args.etag_files

    except AttributeError:
        etag_files = None

    if etag_files:
        r = [etag(file) for file in etag_files]
        if len(r) == 1:
            print(r[0])
        else:
            print(' '.join(r))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from qiniu import config, QiniuMacAuth
from qiniu import http
from qiniu.utils import urlsafe_base64_encode, entry, decode_entry
from qiniu.http.endpoint import Endpoint
from qiniu.http.region import Region, ServiceName
from qiniu.http.regions_provider import get_default_regions_provider

from ._bucket_default_retrier import get_default_retrier


class BucketManager(object):
    """空间管理类

    主要涉及了空间资源管理及批量操作接口的实现，具体的接口规格可以参考：
    https://developer.qiniu.com/kodo/1274/rs

    Attributes:
        auth: 账号管理密钥对，Auth对象
    """

    def __init__(
        self,
        auth,
        zone=None,
        regions=None,
        query_regions_endpoints=None,
        preferred_scheme='http'
    ):
        """
        Parameters
        ----------
        auth: Auth
        zone: LegacyRegion
        regions: list[Region]
        query_regions_endpoints: list[Endpoint]
        preferred_scheme: str, default='http'
        """
        self.auth = auth
        self.mac_auth = QiniuMacAuth(
            auth.get_access_key(),
            auth.get_secret_key(),
            auth.disable_qiniu_timestamp_signature)

        if zone is None:
            self.zone = config.get_default('default_zone')
        else:
            self.zone = zone

        self.regions = regions
        self.query_regions_endpoints = query_regions_endpoints
        self.preferred_scheme = preferred_scheme

    def list(self, bucket, prefix=None, marker=None, limit=None, delimiter=None):
        """前缀查询:

        1. 首次请求 marker = None
        2. 无论 err 值如何，均应该先看 ret.get('items') 是否有内容
        3. 如果后续没有更多数据，err 返回 EOF，marker 返回 None（但不通过该特征来判断是否结束）
        具体规格参考:
        https://developer.qiniu.com/kodo/api/list

        Args:
            bucket:     空间名
            prefix:     列举前缀
            marker:     列举标识符
            limit:      单次列举个数限制
            delimiter:  指定目录分隔符

        Returns:
            一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
            一个ResponseInfo对象
            一个EOF信息。
        """
        options = {
            'bucket': bucket,
        }
        if marker is not None:
            options['marker'] = marker
        if limit is not None:
            options['limit'] = limit
        if prefix is not None:
            options['prefix'] = prefix
        if delimiter is not None:
            options['delimiter'] = delimiter

        ret, info = self.__server_do_with_retrier(
            bucket,
            [ServiceName.RSF],
            '/list',
            data=options,
            method='GET'
        )

        eof = False
        if ret and not ret.get('marker'):
            eof = True

        return ret, eof, info

    def list_domains(self, bucket):
        """获取 Bucket 空间域名
        https://developer.qiniu.com/kodo/3949/get-the-bucket-space-domain

        Args:
            bucket: 空间名

        Returns:
            resBody, respInfo
            resBody 为绑定的域名列表，格式：["example.com"]
        """
        return self.__uc_do_with_retrier('/v2/domains?tbl={0}'.format(bucket))

    def stat(self, bucket, key):
        """获取文件信息:

        获取资源的元信息，但不返回文件内容，具体规格参考：
        https://developer.qiniu.com/kodo/api/1308/stat

        Args:
            bucket: 待获取信息资源所在的空间
            key:    待获取资源的文件名

        Returns:
            一个dict变量，类似：
                {
                    "fsize":        5122935,
                    "hash":         "ljfockr0lOil_bZfyaI2ZY78HWoH",
                    "mimeType":     "application/octet-stream",
                    "putTime":      13603956734587420
                    "type":         0
                }
            一个ResponseInfo对象
        """
        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/stat/{0}'.format(resource)
        )

    def delete(self, bucket, key):
        """删除文件:

        删除指定资源，具体规格参考：
        https://developer.qiniu.com/kodo/api/delete

        Args:
            bucket: 待获取信息资源所在的空间
            key:    待获取资源的文件名

        Returns:
            一个dict变量，成功返回NULL，失败返回{"error": "<errMsg string>"}
            一个ResponseInfo对象
        """
        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/delete/{0}'.format(resource)
        )

    def rename(self, bucket, key, key_to, force='false'):
        """重命名文件:

        给资源进行重命名，本质为move操作。

        Args:
            bucket: 待操作资源所在空间
            key:    待操作资源文件名
            key_to: 目标资源文件名
            force:  是否强制覆盖

        Returns:
            一个dict变量，成功返回NULL，失败返回{"error": "<errMsg string>"}
            一个ResponseInfo对象
        """
        return self.move(bucket, key, bucket, key_to, force)

    def move(self, bucket, key, bucket_to, key_to, force='false'):
        """移动文件:

        将资源从一个空间到另一个空间，具体规格参考：
        https://developer.qiniu.com/kodo/api/move

        Args:
            bucket:     待操作资源所在空间
            bucket_to:  目标资源空间名
            key:        待操作资源文件名
            key_to:     目标资源文件名
            force:      是否强制覆盖

        Returns:
            一个dict变量，成功返回NULL，失败返回{"error": "<errMsg string>"}
            一个ResponseInfo对象
        """
        src = entry(bucket, key)
        dst = entry(bucket_to, key_to)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/move/{src}/{dst}/force/{force}'.format(
                src=src,
                dst=dst,
                force=force
            )
        )

    def copy(self, bucket, key, bucket_to, key_to, force='false'):
        """复制文件:

        将指定资源复制为新命名资源，具体规格参考：
        https://developer.qiniu.com/kodo/api/copy

        Args:
            bucket:     待操作资源所在空间
            bucket_to:  目标资源空间名
            key:        待操作资源文件名
            key_to:     目标资源文件名
            force:      是否强制覆盖

        Returns:
            一个dict变量，成功返回NULL，失败返回{"error": "<errMsg string>"}
            一个ResponseInfo对象
        """
        src = entry(bucket, key)
        dst = entry(bucket_to, key_to)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/copy/{src}/{dst}/force/{force}'.format(
                src=src,
                dst=dst,
                force=force
            )
        )

    def fetch(self, url, bucket, key=None, hostscache_dir=None):
        """抓取文件:
        从指定URL抓取资源，并将该资源存储到指定空间中，具体规格参考：
        https://developer.qiniu.com/kodo/api/fetch

        Args:
            url:      指定的URL
            bucket:   目标资源空间
            key:      目标资源文件名
            hostscache_dir: deprecated, 此参数不再生效，可修改 get_default_regions_provider 返回对象的属性达成同样功能；
                查询区域缓存文件保存位置

        Returns:
            一个dict变量：
                成功 返回{'fsize': <fsize int>, 'hash': <hash string>, 'key': <key string>, 'mimeType': <mimeType string>}
                失败 返回 None
            一个ResponseInfo对象
        """
        resource = urlsafe_base64_encode(url)
        to = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.IO],
            '/fetch/{0}/to/{1}'.format(resource, to)
        )

    def prefetch(self, bucket, key, hostscache_dir=None):
        """镜像回源预取文件:

        从镜像源站抓取资源到空间中，如果空间中已经存在，则覆盖该资源，具体规格参考
        https://developer.qiniu.com/kodo/api/prefetch

        Args:
            bucket: 待获取资源所在的空间
            key:    代获取资源文件名
            hostscache_dir: deprecated, 此参数不再生效，可修改 get_default_regions_provider 返回对象的属性达成同样功能；
                查询区域缓存文件保存位置

        Returns:
            一个dict变量，成功返回NULL，失败返回{"error": "<errMsg string>"}
            一个ResponseInfo对象
        """
        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.IO],
            '/prefetch/{0}'.format(resource)
        )

    def change_mime(self, bucket, key, mime):
        """修改文件mimeType:

        主动修改指定资源的文件类型，具体规格参考：
        https://developer.qiniu.com/kodo/api/chgm

        Args:
            bucket: 待操作资源所在空间
            key:    待操作资源文件名
            mime:   待操作文件目标mimeType
        """
        resource = entry(bucket, key)
        encode_mime = urlsafe_base64_encode(mime)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/chgm/{0}/mime/{1}'.format(resource, encode_mime)
        )

    def change_type(self, bucket, key, storage_type):
        """修改文件的存储类型

        修改文件的存储类型，参考文档：
        https://developer.qiniu.com/kodo/3710/chtype

        Args:
            bucket:         待操作资源所在空间
            key:            待操作资源文件名
            storage_type:   待操作资源存储类型，0为普通存储，1为低频存储，2 为归档存储，3 为深度归档，4 为归档直读存储
        """
        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/chtype/{0}/type/{1}'.format(resource, storage_type)
        )

    def restoreAr(self, bucket, key, freezeAfter_days):
        """
        restore_ar 的别名，用于兼容旧版本

        Args:
            bucket:         待操作资源所在空间
            key:            待操作资源文件名
            freezeAfter_days:   解冻有效时长，取值范围 1～7
        """
        return self.restore_ar(
            bucket,
            key,
            freezeAfter_days
        )

    def restore_ar(self, bucket, key, freeze_after_days):
        """
        解冻归档存储、深度归档存储文件

        对归档存储、深度归档存储文件，进行解冻操作参考文档：
        https://developer.qiniu.com/kodo/6380/restore-archive

        Parameters
        ----------
        bucket: str
        key: str
        freeze_after_days: int

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """

        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/restoreAr/{0}/freezeAfterDays/{1}'.format(resource, freeze_after_days)
        )

    def change_status(self, bucket, key, status, cond):
        """修改文件的状态

        修改文件的存储类型为可用或禁用：

        Args:
            bucket:         待操作资源所在空间
            key:            待操作资源文件名
            status:   待操作资源存储类型，0为启用，1为禁用
        """
        resource = entry(bucket, key)
        url_resource = '/chstatus/{0}/status/{1}'.format(resource, status)
        if cond and isinstance(cond, dict):
            condstr = urlsafe_base64_encode(
                '&'.join(
                    '='.join([k, v])
                    for k, v in cond.items()
                )
            )
            url_resource += '/cond/{0}'.format(condstr)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            url_resource
        )

    def set_object_lifecycle(
        self,
        bucket,
        key,
        to_line_after_days=0,
        to_archive_after_days=0,
        to_deep_archive_after_days=0,
        delete_after_days=0,
        cond=None,
        to_archive_ir_after_days=0
    ):
        """

        设置对象的生命周期

        Args:
            bucket: 目标空间
            key: 目标资源
            to_line_after_days: 多少天后将文件转为低频存储，设置为 -1 表示取消已设置的转低频存储的生命周期规则， 0 表示不修改转低频生命周期规则。
            to_archive_after_days: 多少天后将文件转为归档存储，设置为 -1 表示取消已设置的转归档存储的生命周期规则， 0 表示不修改转归档生命周期规则。
            to_deep_archive_after_days: 多少天后将文件转为深度归档存储，设置为 -1 表示取消已设置的转深度归档存储的生命周期规则， 0 表示不修改转深度归档生命周期规则
            delete_after_days: 多少天后将文件删除，设置为 -1 表示取消已设置的删除存储的生命周期规则， 0 表示不修改删除存储的生命周期规则。
            cond: 匹配条件，只有条件匹配才会设置成功，当前支持设置 hash、mime、fsize、putTime。
            to_archive_ir_after_days: 多少天后将文件转为归档直读存储，设置为 -1 表示取消已设置的转归档只读存储的生命周期规则， 0 表示不修改转归档只读存储生命周期规则。

        Returns:
            resBody, respInfo

        """
        options = [
            'toIAAfterDays', str(to_line_after_days),
            'toArchiveIRAfterDays', str(to_archive_ir_after_days),
            'toArchiveAfterDays', str(to_archive_after_days),
            'toDeepArchiveAfterDays', str(to_deep_archive_after_days),
            'deleteAfterDays', str(delete_after_days)
        ]
        if cond and isinstance(cond, dict):
            cond_str = '&'.join(
                '='.join([k, v])
                for k, v in cond.items()
            )
            options += ['cond', urlsafe_base64_encode(cond_str)]
        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            service_names=[ServiceName.RS],
            url_resource='/lifecycle/{0}/{1}'.format(resource, '/'.join(options)),
        )

    def batch(self, operations):
        """批量操作:

        在单次请求中进行多个资源管理操作，具体规格参考：
        https://developer.qiniu.com/kodo/api/batch

        Args:
            operations: 资源管理操作数组，可通过

        Returns:
            一个dict变量，返回结果类似：
                [
                    { "code": <HttpCode int>, "data": <Data> },
                    { "code": <HttpCode int> },
                    { "code": <HttpCode int> },
                    { "code": <HttpCode int> },
                    { "code": <HttpCode int>, "data": { "error": "<ErrorMessage string>" } },
                    ...
                ]
            一个ResponseInfo对象
        """
        if not operations:
            # change to ValueError when make break changes version
            raise Exception('operations is empty')
        bucket = ''
        for op in operations:
            segments = op.split('/')
            e = segments[1] if len(segments) >= 2 else ''
            bucket, _ = decode_entry(e)
            if bucket:
                break
        if not bucket:
            # change to ValueError when make break changes version
            raise Exception('bucket is empty')

        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/batch',
            {'op': operations}
        )

    def buckets(self):
        """获取所有空间名:

        获取指定账号下所有的空间名。

        Returns:
            一个dict变量，类似：
                [ <Bucket1>, <Bucket2>, ... ]
            一个ResponseInfo对象
        """
        return self.__uc_do_with_retrier('/buckets')

    def delete_after_days(self, bucket, key, days):
        """更新文件生命周期

        Returns:
            一个dict变量，返回结果类似：
                [
                    { "code": <HttpCode int>, "data": <Data> },
                    { "code": <HttpCode int> },
                    { "code": <HttpCode int> },
                    { "code": <HttpCode int> },
                    { "code": <HttpCode int>, "data": { "error": "<ErrorMessage string>" } },
                    ...
                ]
            一个ResponseInfo对象
        Args:
            bucket: 目标资源空间
            key:    目标资源文件名
            days:   指定天数
        """
        resource = entry(bucket, key)
        return self.__server_do_with_retrier(
            bucket,
            [ServiceName.RS],
            '/deleteAfterDays/{0}/{1}'.format(resource, days)
        )

    def mkbucketv3(self, bucket_name, region):
        """
        创建存储空间，全局唯一，其他账号有同名空间就无法创建

        Args:
            bucket_name: 存储空间名
            region: 存储区域
        """
        return self.__uc_do_with_retrier(
            '/mkbucketv3/{0}/region/{1}'.format(bucket_name, region)
        )

    def list_bucket(self, region):
        """
        列举存储空间列表

        Args:
        """
        return self.__uc_do_with_retrier('/v3/buckets?region={0}'.format(region))

    def bucket_info(self, bucket_name):
        """
        获取存储空间信息

        Args:
            bucket_name: 存储空间名
        """
        return self.__uc_do_with_retrier('/v2/bucketInfo?bucket={0}'.format(bucket_name))

    def bucket_domain(self, bucket_name):
        """
        获取存储空间域名列表
        Args:
            bucket_name: 存储空间名
        """
        return self.list_domains(bucket_name)

    def change_bucket_permission(self, bucket_name, private):
        """
        设置 存储空间访问权限
        https://developer.qiniu.com/kodo/api/3946/set-bucket-private
        Args:
            bucket_name: 存储空间名
            private: 0 公开；1 私有 ,str类型
        """
        return self.__uc_do_with_retrier(
            '/private?bucket={0}&private={1}'.format(bucket_name, private)
        )

    def _get_regions_provider(self, bucket_name):
        """
        Parameters
        ----------
        bucket_name: str

        Returns
        -------
        Iterable[Region]
        """
        if self.regions:
            return self.regions

        # handle compatibility for legacy config
        if self.zone and any(
            hasattr(self.zone, attr_name) and getattr(self.zone, attr_name)
            for attr_name in [
                'io_host',
                'rs_host',
                'rsf_host',
                'api_host'
            ]
        ):
            return [self.zone]

        # handle compatibility for default_query_region_host
        query_regions_endpoints = self.query_regions_endpoints
        if not query_regions_endpoints:
            query_region_host = config.get_default('default_query_region_host')
            query_region_backup_hosts = config.get_default('default_query_region_backup_hosts')
            query_regions_endpoints = [
                Endpoint.from_host(h)
                for h in [query_region_host] + query_region_backup_hosts
            ]

        return get_default_regions_provider(
            query_endpoints_provider=query_regions_endpoints,
            access_key=self.auth.get_access_key(),
            bucket_name=bucket_name,
            preferred_scheme=self.preferred_scheme
        )

    def __uc_do_with_retrier(self, url_resource, data=None):
        """
        Parameters
        ----------
        url_resource: url
        data: dict or None

        Returns
        -------
        ret: dict or None
        resp: ResponseInfo
        """
        regions = self.regions

        # ignore self.zone by no uc in it
        # handle compatibility for default_uc
        if not regions:
            uc_host = config.get_default('default_uc_host')
            uc_backup_hosts = config.get_default('default_uc_backup_hosts')
            uc_endpoints = [
                Endpoint.from_host(h)
                for h in [uc_host] + uc_backup_hosts
            ]
            regions = [Region(services={ServiceName.UC: uc_endpoints})]

        retrier = get_default_retrier(
            regions_provider=regions,
            service_names=[ServiceName.UC]
        )

        attempt = None
        for attempt in retrier:
            with attempt:
                host = attempt.context.get('endpoint').get_value(scheme=self.preferred_scheme)
                url = host + url_resource
                attempt.result = self.__post(url, data)
                ret, resp = attempt.result
                if resp.ok() and ret:
                    return attempt.result
                if not resp.need_retry():
                    return attempt.result

        if attempt is None:
            raise RuntimeError('Retrier is not working. attempt is None')

        return attempt.result

    def __server_do_with_retrier(self, bucket_name, service_names, url_resource, data=None, method='POST'):
        """
        Parameters
        ----------
        bucket_name: str
        service_names: List[ServiceName]
        url_resource: str
        data: dict or None
        method: str

        Returns
        -------
        ret: dict or None
        resp: ResponseInfo
        """
        if not service_names:
            raise ValueError('service_names is empty')

        retrier = get_default_retrier(
            regions_provider=self._get_regions_provider(bucket_name=bucket_name),
            service_names=service_names
        )

        method = method.upper()
        if method == 'POST':
            send_request = self.__post
        elif method == 'GET':
            send_request = self.__get
        else:
            raise ValueError('"method" must be "POST" or "GET"')

        attempt = None
        for attempt in retrier:
            with attempt:
                host = attempt.context.get('endpoint').get_value(scheme=self.preferred_scheme)
                url = host + url_resource
                attempt.result = send_request(url, data)
                ret, resp = attempt.result
                if resp.ok() and ret:
                    return attempt.result
                if not resp.need_retry():
                    return attempt.result

        if attempt is None:
            raise RuntimeError('Retrier is not working. attempt is None')

        return attempt.result

    def __post(self, url, data=None):
        return http._post_with_qiniu_mac(url, data, self.mac_auth)

    def __get(self, url, params=None):
        return http._get_with_qiniu_mac(url, params, self.mac_auth)


def _build_op(*args):
    return '/'.join(map(str, args))


def build_batch_copy(source_bucket, key_pairs, target_bucket, force='false'):
    """
    Parameters
    ----------
    source_bucket: str
    key_pairs: dict
    target_bucket: str
    force: str

    Returns
    -------
    list[str]
    """
    return _two_key_batch('copy', source_bucket, key_pairs, target_bucket, force)


def build_batch_rename(bucket, key_pairs, force='false'):
    """
    Parameters
    ----------
    bucket: str
    key_pairs: dict
    force: str

    Returns
    -------
    list[str]
    """
    return build_batch_move(bucket, key_pairs, bucket, force)


def build_batch_move(source_bucket, key_pairs, target_bucket, force='false'):
    """
    Parameters
    ----------
    source_bucket: str
    key_pairs: dict
    target_bucket: str
    force: str

    Returns
    -------
    list[str]
    """
    return _two_key_batch('move', source_bucket, key_pairs, target_bucket, force)


def build_batch_restoreAr(bucket, keys):
    """
    alias for build_batch_restore_ar for compatibility with old version

    Parameters
    ----------
    bucket: str
    keys: dict

    Returns
    -------
    list[str]
    """
    return build_batch_restore_ar(bucket, keys)


def build_batch_restore_ar(bucket, keys):
    """
    Parameters
    ----------
    bucket: str
    keys: dict

    Returns
    -------
    list[str]
    """
    keys = {
        k: ['freezeAfterDays', v]
        for k, v in keys.items()
    }
    return _one_key_batch('restoreAr', bucket, keys)


def build_batch_delete(bucket, keys):
    """
    Parameters
    ----------
    bucket: str
    keys: list[str]

    Returns
    -------
    list[str]
    """
    return _one_key_batch('delete', bucket, keys)


def build_batch_stat(bucket, keys):
    """
    Parameters
    ----------
    bucket: str
    keys: list[str]

    Returns
    -------
    list[str]
    """
    return _one_key_batch('stat', bucket, keys)


def _one_key_batch(operation, bucket, keys):
    """
    Parameters
    ----------
    operation: str
    bucket: str
    keys: list[str] or dict

    Returns
    -------
    list[str]
    """
    # use functools.singledispatch to refactor when min version of python >= 3.4
    if isinstance(keys, list):
        return [
            _build_op(
                operation,
                entry(bucket, key),
            )
            for key in keys
        ]
    elif isinstance(keys, dict):
        return [
            _build_op(
                operation,
                entry(bucket, key),
                *opts
            )
            for key, opts in keys.items()
        ]
    else:
        raise TypeError('"keys" only support list or dict')


def _two_key_batch(operation, source_bucket, key_pairs, target_bucket=None, force='false'):
    """

    Parameters
    ----------
    operation: str
    source_bucket: str
    key_pairs: dict
    target_bucket: str
    force: str

    Returns
    -------
    list[str]
    """
    if target_bucket is None:
        target_bucket = source_bucket
    return _one_key_batch(
        operation,
        source_bucket,
        {
            src_key: [
                entry(target_bucket, dst_key),
                'force',
                force
            ]
            for src_key, dst_key in key_pairs.items()
        }
    )


def _three_key_batch(operation, bucket, keys):
    """
    .. deprecated: Use `_one_key_batch` instead.
        `keys` could be `{key: [freezeAfterDays, days]}`

    Parameters
    ----------
    operation: str
    bucket: str
    keys: dict

    Returns
    -------
    list[str]
    """
    keys = {
        k: ['freezeAfterDays', v]
        for k, v in keys.items()
    }
    return _one_key_batch(operation, bucket, keys)



# -*- coding: utf-8 -*-
import base64
from datetime import datetime
import hmac
import os
import time
from hashlib import sha1
from requests.auth import AuthBase
from .compat import urlparse, json, b
from .utils import urlsafe_base64_encode, canonical_mime_header_key

# 上传策略，参数规格详见
# https://developer.qiniu.com/kodo/manual/1206/put-policy
# the `str()` prevent implicit concatenation of string. DON'T remove it.
# for example, avoid you lost comma at the end of line in middle.
_policy_fields = {
    str('callbackUrl'),  # 回调URL
    str('callbackBody'),  # 回调Body
    str('callbackHost'),  # 回调URL指定的Host
    str('callbackBodyType'),  # 回调Body的Content-Type
    str('callbackFetchKey'),  # 回调FetchKey模式开关

    str('returnUrl'),  # 上传端的303跳转URL
    str('returnBody'),  # 上传端简单反馈获取的Body

    str('endUser'),  # 回调时上传端标识
    str('saveKey'),  # 自定义资源名
    str('forceSaveKey'),  # saveKey的优先级设置。为 true 时，saveKey不能为空，会忽略客户端指定的key，强制使用saveKey进行文件命名。参数不设置时，默认值为false
    str('insertOnly'),  # 插入模式开关

    str('detectMime'),  # MimeType侦测开关
    str('mimeLimit'),  # MimeType限制
    str('fsizeLimit'),  # 上传文件大小限制
    str('fsizeMin'),  # 上传文件最少字节数
    str('keylimit'),  # 设置允许上传的key列表，字符串数组类型，数组长度不可超过20个，如果设置了这个字段，上传时必须提供key

    str('persistentOps'),  # 持久化处理操作，与 persistentWorkflowTemplateID 二选一
    str('persistentNotifyUrl'),  # 持久化处理结果通知URL
    str('persistentPipeline'),  # 持久化处理独享队列
    str('persistentType'),  # 为 `1` 时，开启闲时任务，必须是 int 类型
    str('persistentWorkflowTemplateID'),  # 工作流模板 ID，与 persistentOps 二选一

    str('deleteAfterDays'),  # 文件多少天后自动删除
    str('fileType'),  # 文件的存储类型，0为标准存储，1为低频存储，2为归档存储，3为深度归档存储，4为归档直读存储
    str('isPrefixalScope'),  # 指定上传文件必须使用的前缀

    str('transform'),  # deprecated
    str('transformFallbackKey'),  # deprecated
    str('transformFallbackMode'),  # deprecated
}


class Auth(object):
    """七牛安全机制类

    该类主要内容是七牛上传凭证、下载凭证、管理凭证三种凭证的签名接口的实现，以及回调验证。

    Attributes:
        __access_key: 账号密钥对中的accessKey，详见 https://portal.qiniu.com/user/key
        __secret_key: 账号密钥对重的secretKey，详见 https://portal.qiniu.com/user/key
    """

    def __init__(self, access_key, secret_key, disable_qiniu_timestamp_signature=None):
        """初始化Auth类"""
        self.__checkKey(access_key, secret_key)
        self.__access_key = access_key
        self.__secret_key = b(secret_key)
        self.disable_qiniu_timestamp_signature = disable_qiniu_timestamp_signature

    def get_access_key(self):
        return self.__access_key

    def get_secret_key(self):
        return self.__secret_key

    def __token(self, data):
        data = b(data)
        hashed = hmac.new(self.__secret_key, data, sha1)
        return urlsafe_base64_encode(hashed.digest())

    def token(self, data):
        return '{0}:{1}'.format(self.__access_key, self.__token(data))

    def token_with_data(self, data):
        data = urlsafe_base64_encode(data)
        return '{0}:{1}:{2}'.format(
            self.__access_key, self.__token(data), data)

    def token_of_request(self, url, body=None, content_type=None):
        """带请求体的签名（本质上是管理凭证的签名）

        Args:
            url:          待签名请求的url
            body:         待签名请求的body
            content_type: 待签名请求的body的Content-Type

        Returns:
            管理凭证
        """
        parsed_url = urlparse(url)
        query = parsed_url.query
        path = parsed_url.path
        data = path
        if query != '':
            data = ''.join([data, '?', query])
        data = ''.join([data, "\n"])

        if body:
            mimes = [
                'application/x-www-form-urlencoded'
            ]
            if content_type in mimes:
                data += body

        return '{0}:{1}'.format(self.__access_key, self.__token(data))

    @staticmethod
    def __checkKey(access_key, secret_key):
        if not (access_key and secret_key):
            raise ValueError('invalid key')

    def private_download_url(self, url, expires=3600):
        """生成私有资源下载链接

        Args:
            url:     私有空间资源的原始URL
            expires: 下载凭证有效期，默认为3600s

        Returns:
            私有资源的下载链接
        """
        deadline = int(time.time()) + expires
        if '?' in url:
            url += '&'
        else:
            url += '?'
        url = '{0}e={1}'.format(url, str(deadline))

        token = self.token(url)
        return '{0}&token={1}'.format(url, token)

    def upload_token(
            self,
            bucket,
            key=None,
            expires=3600,
            policy=None,
            strict_policy=True):
        """生成上传凭证

        Args:
            bucket:  上传的空间名
            key:     上传的文件名，默认为空
            expires: 上传凭证的过期时间，默认为3600s
            policy:  上传策略，默认为空
            strict_policy:  严格模式，将校验 policy 字段，默认为 True

        Returns:
            上传凭证
        """
        if bucket is None or bucket == '':
            raise ValueError('invalid bucket name')

        scope = bucket
        if key is not None:
            scope = '{0}:{1}'.format(bucket, key)

        args = dict(
            scope=scope,
            deadline=int(time.time()) + expires,
        )

        if policy is not None:
            self.__copy_policy(policy, args, strict_policy)

        return self.__upload_token(args)

    @staticmethod
    def up_token_decode(up_token):
        up_token_list = up_token.split(':')
        ak = up_token_list[0]
        sign = base64.urlsafe_b64decode(up_token_list[1])
        decode_policy = base64.urlsafe_b64decode(up_token_list[2])
        decode_policy = decode_policy.decode('utf-8')
        dict_policy = json.loads(decode_policy)
        return ak, sign, dict_policy

    @staticmethod
    def get_bucket_name(up_token):
        _, _, policy = Auth.up_token_decode(up_token)
        if not policy or not policy['scope']:
            return None
        return policy['scope'].split(':', 1)[0]

    def __upload_token(self, policy):
        data = json.dumps(policy, separators=(',', ':'))
        return self.token_with_data(data)

    def verify_callback(
        self,
        origin_authorization,
        url,
        body,
        content_type='application/x-www-form-urlencoded',
        method='GET',
        headers=None
    ):
        """
        Qbox 回调验证

        Parameters
        ----------
        origin_authorization: str
            回调时请求 Header 中的 Authorization 字段
        url: str
            回调请求的 url
        body: str
            回调请求的 body
        content_type: str
            回调请求的 Content-Type
        method: str
            回调请求的 method，Qiniu 签名必须传入，默认 GET
        headers: dict
            回调请求的 headers，Qiniu 签名必须传入，默认为空字典

        Returns
        -------
        bool
            返回 True 表示验证成功，返回 False 表示验证失败
        """
        if headers is None:
            headers = {}

        # 兼容 Qiniu 签名
        if origin_authorization.startswith("Qiniu"):
            qn_auth = QiniuMacAuth(
                access_key=self.__access_key,
                secret_key=self.__secret_key,
                disable_qiniu_timestamp_signature=True
            )
            return qn_auth.verify_callback(
                origin_authorization,
                url=url,
                body=body,
                content_type=content_type,
                method=method,
                headers=headers
            )

        token = self.token_of_request(url, body, content_type)
        authorization = 'QBox {0}'.format(token)
        return origin_authorization == authorization

    @staticmethod
    def __copy_policy(policy, to, strict_policy):
        for k, v in policy.items():
            if (not strict_policy) or k in _policy_fields:
                to[k] = v


class RequestsAuth(AuthBase):
    def __init__(self, auth):
        self.auth = auth

    def __call__(self, r):
        if r.body is not None and r.headers['Content-Type'] == 'application/x-www-form-urlencoded':
            token = self.auth.token_of_request(
                r.url, r.body, 'application/x-www-form-urlencoded')
        else:
            token = self.auth.token_of_request(r.url)
        r.headers['Authorization'] = 'QBox {0}'.format(token)
        return r


class QiniuMacAuth(object):
    """
    Sign Requests

    Attributes:
        __access_key
        __secret_key

    https://developer.qiniu.com/kodo/1201/access-token
    """

    def __init__(self, access_key, secret_key, disable_qiniu_timestamp_signature=None):
        self.qiniu_header_prefix = "X-Qiniu-"
        self.__checkKey(access_key, secret_key)
        self.__access_key = access_key
        self.__secret_key = b(secret_key)
        self.disable_qiniu_timestamp_signature = disable_qiniu_timestamp_signature

    def __token(self, data):
        data = b(data)
        hashed = hmac.new(self.__secret_key, data, sha1)
        return urlsafe_base64_encode(hashed.digest())

    @property
    def should_sign_with_timestamp(self):
        if self.disable_qiniu_timestamp_signature is not None:
            return not self.disable_qiniu_timestamp_signature
        if os.getenv('DISABLE_QINIU_TIMESTAMP_SIGNATURE', '').lower() == 'true':
            return False
        return True

    def token_of_request(
            self,
            method,
            host,
            url,
            qheaders,
            content_type=None,
            body=None):
        """
        <Method> <PathWithRawQuery>
        Host: <Host>
        Content-Type: <ContentType>
        [<X-Qiniu-*> Headers]

        [<Body>] #这里的 <Body> 只有在 <ContentType> 存在且不为 application/octet-stream 时才签进去。

        """
        parsed_url = urlparse(url)
        netloc = parsed_url.netloc
        path = parsed_url.path
        query = parsed_url.query

        if not host:
            host = netloc

        path_with_query = path
        if query != '':
            path_with_query = ''.join([path_with_query, '?', query])
        data = ''.join([
            "%s %s" % (method, path_with_query),
            "\n",
            "Host: %s" % host
        ])

        if content_type:
            data += "\n"
            data += "Content-Type: %s" % content_type

        if qheaders:
            data += "\n"
            data += qheaders

        data += "\n\n"

        if content_type and content_type != "application/octet-stream" and body:
            if isinstance(body, bytes):
                data += body.decode(encoding='UTF-8')
            else:
                data += body
        return '{0}:{1}'.format(self.__access_key, self.__token(data))

    def qiniu_headers(self, headers):
        qiniu_fields = [
            key for key in headers
            if key.startswith(self.qiniu_header_prefix) and len(key) > len(self.qiniu_header_prefix)
        ]
        return '\n'.join([
            '%s: %s' % (canonical_mime_header_key(key), headers.get(key)) for key in sorted(qiniu_fields)
        ])

    def verify_callback(
        self,
        origin_authorization,
        url,
        body,
        content_type='application/x-www-form-urlencoded',
        method='GET',
        headers=None
    ):
        """
        Qiniu 回调验证

        Parameters
        ----------
        origin_authorization: str
            回调时请求 Header 中的 Authorization 字段
        url: str
            回调请求的 url
        body: str
            回调请求的 body
        content_type: str
            回调请求的 Content-Type
        method: str
            回调请求的 Method
        headers: dict
            回调请求的 headers

        Returns
        -------

        """
        if headers is None:
            headers = {}
        token = self.token_of_request(
            method=method,
            host=headers.get('Host', None),
            url=url,
            qheaders=self.qiniu_headers(headers),
            content_type=content_type,
            body=body
        )
        authorization = 'Qiniu {0}'.format(token)
        return origin_authorization == authorization

    @staticmethod
    def __checkKey(access_key, secret_key):
        if not (access_key and secret_key):
            raise ValueError('QiniuMacAuthSign : Invalid key')


class QiniuMacRequestsAuth(AuthBase):
    """
    Attributes:
        auth (QiniuMacAuth):
    """
    def __init__(self, auth):
        """
        Args:
            auth (QiniuMacAuth):
        """
        self.auth = auth

    def __call__(self, r):
        if r.headers.get('Content-Type', None) is None:
            r.headers['Content-Type'] = 'application/x-www-form-urlencoded'

        if self.auth.should_sign_with_timestamp:
            x_qiniu_date = datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')
            r.headers['X-Qiniu-Date'] = x_qiniu_date

        token = self.auth.token_of_request(
            r.method,
            r.headers.get('Host', None),
            r.url,
            self.auth.qiniu_headers(r.headers),
            r.headers.get('Content-Type', None),
            r.body
        )
        r.headers['Authorization'] = 'Qiniu {0}'.format(token)
        return r



import abc

import qiniu.config as config
from qiniu.region import LegacyRegion
from qiniu.http.endpoint import Endpoint
from qiniu.http.regions_provider import get_default_regions_provider

# type import
from qiniu.auth import Auth # noqa
from qiniu.http.region import Region, ServiceName  # noqa


class UploaderBase(object):
    """
    Attributes
    ----------
    bucket_name: str
    auth: Auth
    regions: list[Region]
    """
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        bucket_name,
        **kwargs
    ):
        """
        Parameters
        ----------
        bucket_name: str
            The name of bucket which you want to upload to.
        auth: Auth
            The instance of Auth to sign requests.
        regions: list[Region], default=[]
            The regions of bucket. It will be queried if not specified.
        kwargs
            The others arguments may be used by subclass.
        """
        # default bucket_name
        self.bucket_name = bucket_name

        # change the default when implements AuthProvider
        self.auth = kwargs.get('auth', None)

        # regions config
        regions = kwargs.get('regions', None)
        if not regions:
            regions = []
        self.regions = regions

        query_regions_endpoints = kwargs.get('query_regions_endpoints', None)
        if not query_regions_endpoints:
            query_regions_endpoints = []
        self.query_regions_endpoints = query_regions_endpoints

        self.preferred_scheme = kwargs.get('preferred_scheme', 'http')

        # change the default value to False when remove config.get_default('default_zone')
        self.accelerate_uploading = kwargs.get('accelerate_uploading', None)

    def get_up_token(
        self,
        bucket_name=None,
        key=None,
        expired=None,
        policy=None,
        strict_policy=None,
        **_kwargs
    ):
        """
        Generate up token

        Parameters
        ----------
        bucket_name: str
        key: str
        expired: int
            seconds
        policy: dict
        strict_policy: bool
        _kwargs: dict
            useless for now, just for compatibility

        Returns
        -------
        str
        """
        if not self.auth:
            raise ValueError('can not get up_token by auth not provided')

        bucket_name = bucket_name if bucket_name else self.bucket_name

        kwargs_for_up_token = {
            k: v
            for k, v in {
                'bucket': bucket_name,
                'key': key,
                'expired': expired,
                'policy': policy,
                'strict_policy': strict_policy
            }.items()
            if k
        }
        up_token = self.auth.upload_token(**kwargs_for_up_token)
        return up_token

    def _get_regions_provider(self, access_key=None, bucket_name=None):
        """
        Parameters
        ----------
        access_key: str
        bucket_name: str

        Returns
        -------
        Iterable[Region or LegacyRegion]
        """
        if self.regions:
            return self.regions

        # handle compatibility for default_zone
        if config.is_customized_default('default_zone'):
            return [config.get_default('default_zone')]

        # handle compatibility for default_query_region_host
        query_regions_endpoints = self.query_regions_endpoints
        if not query_regions_endpoints:
            query_region_host = config.get_default('default_query_region_host')
            query_region_backup_hosts = config.get_default('default_query_region_backup_hosts')
            query_regions_endpoints = [
                Endpoint.from_host(h)
                for h in [query_region_host] + query_region_backup_hosts
            ]

        # get regions from default regions provider
        if not self.auth and not access_key:
            raise ValueError('Must provide access_key and bucket_name if auth is unavailable.')
        if not access_key:
            access_key = self.auth.get_access_key()
        if not bucket_name:
            bucket_name = self.bucket_name

        return get_default_regions_provider(
            query_endpoints_provider=query_regions_endpoints,
            access_key=access_key,
            bucket_name=bucket_name,
            accelerate_uploading=self.accelerate_uploading,
            preferred_scheme=self.preferred_scheme,
        )

    def _get_regions(self, access_key=None, bucket_name=None):
        """
        .. deprecated::
            This has been deprecated by implemented regions provider and endpoints

        Parameters
        ----------
        access_key: str
        bucket_name: str

        Returns
        -------
        list[LegacyRegion]
        """
        def get_legacy_region(r):
            if isinstance(r, LegacyRegion):
                return r
            opts = {
                'scheme': self.preferred_scheme,
                'accelerate_uploading': self.accelerate_uploading
            }
            if r.services[ServiceName.UP]:
                opts['up_host'] = r.services[ServiceName.UP][0].get_value(self.preferred_scheme)
            if len(r.services[ServiceName.UP]) > 1:
                opts['up_host_backup'] = [
                    e.get_value(self.preferred_scheme)
                    for e in r.services[ServiceName.UP][1:]
                ]
            if r.services[ServiceName.IO]:
                opts['io_host'] = r.services[ServiceName.IO][0].get_value(self.preferred_scheme)
            if r.services[ServiceName.RS]:
                opts['rs_host'] = r.services[ServiceName.RS][0].get_value(self.preferred_scheme)
            if r.services[ServiceName.RSF]:
                opts['rsf_host'] = r.services[ServiceName.RSF][0].get_value(self.preferred_scheme)
            if r.services[ServiceName.API]:
                opts['api_host'] = r.services[ServiceName.API][0].get_value(self.preferred_scheme)
            result = LegacyRegion(**opts)
            result.services = r.services
            result.region_id = r.region_id
            result.s3_region_id = r.s3_region_id
            result.ttl = r.ttl
            result.create_time = r.create_time
            return result

        return [
            get_legacy_region(r)
            for r in self._get_regions_provider(access_key, bucket_name)
        ]

    def _get_up_hosts(self, access_key=None, bucket_name=None):
        """
        get hosts of upload by access key or the first region

        .. deprecated::
            This has been deprecated by implemented regions provider and endpoints

        Returns
        -------
        list[str]
        """
        if not bucket_name:
            bucket_name = self.bucket_name
        if not self.auth and not access_key:
            raise ValueError('Must provide access_key if auth is unavailable.')
        if not access_key:
            access_key = self.auth.get_access_key()

        regions = self._get_regions(access_key, bucket_name)

        if not regions:
            raise ValueError('No region available.')

        # get up hosts in region
        service_names = [ServiceName.UP]
        if self.accelerate_uploading:
            service_names.insert(0, ServiceName.UP_ACC)

        return [
            e.get_value()
            for sn in service_names
            for e in regions[0].services[sn]
        ]

    @abc.abstractmethod
    def upload(
        self,
        key,
        file_path,
        data,
        data_size,
        modify_time,

        part_size,
        mime_type,
        metadata,
        file_name,
        custom_vars,
        **kwargs
    ):
        """
        Upload method

        Parameters
        ----------
        key: str
        file_path: str
        data: IOBase
        data_size: int
        modify_time: int

        part_size: int
        mime_type: str
        metadata: dict
        file_name: str
        custom_vars: dict
        kwargs: dict

        Returns
        -------
        ret: dict
            The parsed response body
        info
            The response
        """



import logging
import math
import functools
from collections import namedtuple
from concurrent import futures
from io import BytesIO
from itertools import chain
from os import path
from threading import Lock
from time import time

from qiniu.compat import is_seekable
from qiniu.auth import Auth
from qiniu.http import qn_http_client, ResponseInfo
from qiniu.http.endpoint import Endpoint
from qiniu.utils import b, io_crc32, urlsafe_base64_encode

from ._default_retrier import ProgressRecord, get_default_retrier
from .abc import ResumeUploaderBase
from .io_chunked import IOChunked


class ResumeUploaderV1(ResumeUploaderBase):
    def _recover_from_record(
        self,
        file_name,
        key,
        context
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV1Context

        Returns
        -------
        _ResumeUploadV1Context
        """
        if not isinstance(context, _ResumeUploadV1Context):
            raise TypeError('"context" must be an instance of _ResumeUploadV1Context')

        if not self.upload_progress_recorder or not any([file_name, key]):
            return context

        record = self.upload_progress_recorder.get_upload_record(
            file_name,
            key
        )

        if not record:
            return context

        record_up_hosts = record.get('up_hosts', [])
        record_part_size = record.get('part_size', None)
        record_modify_time = record.get('modify_time', 0)
        record_context = record.get('contexts', [])

        # compat with old sdk(<= v7.11.1)
        if not record_up_hosts or not record_part_size:
            return context

        # filter expired parts
        if record_modify_time != context.modify_time:
            record_context = []
        else:
            now = time()
            record_context = [
                ctx
                for ctx in record_context
                if (
                    ctx.get('expired_at', 0) > now and
                    ctx.get('part_no', None) and
                    ctx.get('ctx', None)
                )
            ]

        # assign to context
        return context._replace(
            up_hosts=record_up_hosts,
            part_size=record_part_size,
            parts=[
                _ResumeUploadV1Part(
                    part_no=p['part_no'],
                    ctx=p['ctx'],
                    expired_at=p['expired_at'],
                )
                for p in record_context
            ],
            resumed=True
        )

    def _set_to_record(
        self,
        file_name,
        key,
        context
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV1Context

        """
        if not self.upload_progress_recorder or not any([file_name, key]):
            return

        record_data = {
            'up_hosts': context.up_hosts,
            'part_size': context.part_size,
            'modify_time': context.modify_time,
            'contexts': [
                {
                    'ctx': part.ctx,
                    'expired_at': part.expired_at,
                    'part_no': part.part_no
                }
                for part in context.parts
            ]
        }
        self.upload_progress_recorder.set_upload_record(
            file_name,
            key,
            data=record_data
        )

    def _try_delete_record(
        self,
        file_name,
        key,
        context=None,
        resp=None
    ):
        """
        Parameters
        ----------
        file_name: str or None
        key: str or None
        context: _ResumeUploadV1Context
        resp: ResponseInfo
        """
        if not self.upload_progress_recorder or not any([file_name, key]):
            return
        if resp and not resp.ok():
            return
        self.upload_progress_recorder.delete_upload_record(file_name, key)

    def _progress_handler(
        self,
        file_name,
        key,
        context,
        uploaded_size,
        total_size
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV1Context
        uploaded_size: int
        total_size: int

        """
        self._set_to_record(file_name, key, context)
        if not callable(self.progress_handler):
            return
        try:
            self.progress_handler(uploaded_size, total_size)
        except Exception as err:
            err.no_need_retry = True
            raise err

    def _initial_context(
        self,
        key,
        file_name,
        modify_time
    ):
        """
        Parameters
        ----------
        key: str
        file_name: str
        modify_time: float

        Returns
        -------
        _ResumeUploadV1Context
        """
        part_size = 4 * (1024 ** 2)
        context = _ResumeUploadV1Context(
            up_hosts=[],
            part_size=part_size,
            parts=[],
            modify_time=modify_time,
            resumed=False
        )

        # try to recover from record
        return self._recover_from_record(
            key=key,
            file_name=file_name,
            context=context
        )

    def initial_parts(
        self,
        up_token,
        key,
        file_path=None,
        data=None,
        data_size=None,
        modify_time=None,
        part_size=None,
        file_name=None,
        up_endpoint=None,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        key: str
        file_path: str or None
        data: str or None
        modify_time: float
        data_size: int
        part_size: None
            useless for v1 by fixed part size
        file_name: str
        up_endpoint: Endpoint

        kwargs: dict

        Returns
        -------
        context: _ResumeUploadV1Context
        resp: None

        """
        # -- check and initial arguments
        # must provide file_path or data
        if not file_path and not data:
            raise TypeError('Must provide one of file_path or data.')
        if file_path and data:
            raise TypeError('Must provide only one of file_path or data.')

        # data must has length
        if not file_path and not data_size:
            raise TypeError('Must provide size if use data.')

        if not modify_time:
            if file_path:
                modify_time = int(path.getmtime(file_path))
            else:
                modify_time = int(time())

        if not file_name and file_path:
            file_name = path.basename(file_path)

        context = self._initial_context(
            key=key,
            file_name=file_name,
            modify_time=modify_time
        )

        if not context.up_hosts and up_endpoint:
            context.up_hosts.extend([up_endpoint.get_value(self.preferred_scheme)])

        if not context.up_hosts:
            access_key, _, _ = Auth.up_token_decode(up_token)
            context.up_hosts.extend(self._get_up_hosts(access_key))

        return context, None

    def upload_parts(
        self,
        up_token,
        data,
        data_size,
        context,
        **kwargs
    ):
        """

        Parameters
        ----------
        up_token: str
        context: _ResumeUploadV1Context
        data
        data_size: int

        kwargs
            key, file_name

        Returns
        -------
        part: _ResumeUploadV1Part
        resp: ResponseInfo

        """
        # initial arguments
        chunk_list = self.gen_chunk_list(
            size=data_size,
            chunk_size=context.part_size,
            uploaded_chunk_no_list=[
                p.part_no for p in context.parts
            ]
        )
        up_hosts = list(context.up_hosts)
        file_name = kwargs.get('file_name', None)
        key = kwargs.get('key', None)

        # initial upload state
        part, resp = None, None
        uploaded_size = context.part_size * len(context.parts)
        if math.ceil(data_size / context.part_size) in [p.part_no for p in context.parts]:
            # if last part has been uploaded, should correct the uploaded size
            uploaded_size += (data_size % context.part_size) - context.part_size
        lock = Lock()

        if not self.concurrent_executor:
            # upload sequentially
            for chunk in chunk_list:
                part, resp = self.__upload_part(
                    data=data,
                    chunk_info=chunk,
                    up_hosts=up_hosts,
                    up_token=up_token,
                    lock=lock
                )
                if not resp.ok():
                    return None, resp
                elif not part:
                    return resp.json(), resp
                context.parts.append(part)
                uploaded_size += chunk.chunk_size
                self._progress_handler(
                    file_name=file_name,
                    key=key,
                    context=context,
                    uploaded_size=uploaded_size,
                    total_size=data_size
                )
        else:
            # upload concurrently
            future_chunk_dict = {}
            for chunk in chunk_list:
                ftr = self.concurrent_executor.submit(
                    self.__upload_part,
                    data=data,
                    chunk_info=chunk,
                    up_hosts=up_hosts,
                    up_token=up_token,
                    lock=lock
                )
                future_chunk_dict[ftr] = chunk

            first_failed_resp = None
            for ftr in futures.as_completed(future_chunk_dict):
                if ftr.cancelled():
                    continue
                elif ftr.exception():
                    # only keep first failed future,
                    # continue instead return to wait running future done.
                    if first_failed_resp:
                        continue
                    first_failed_resp = ResponseInfo(None, ftr.exception())
                    for not_done in filter(lambda f: not f.done(), future_chunk_dict):
                        not_done.cancel()
                else:
                    part, resp = ftr.result()
                    if not part:
                        if not first_failed_resp:
                            first_failed_resp = resp
                            for not_done in filter(lambda f: not f.done(), future_chunk_dict):
                                not_done.cancel()
                    else:
                        context.parts.append(part)
                        uploaded_size += future_chunk_dict[ftr].chunk_size
                        self._progress_handler(
                            file_name=file_name,
                            key=key,
                            context=context,
                            uploaded_size=uploaded_size,
                            total_size=data_size
                        )
            if first_failed_resp:
                if first_failed_resp.ok():
                    # just compat with old sdk. it's ok when crc32 check failed
                    return first_failed_resp.json(), first_failed_resp
                return None, first_failed_resp

        return part, resp

    def complete_parts(
        self,
        up_token,
        data_size,
        context,
        **kwargs
    ):
        """

        Parameters
        ----------
        up_token: str
        data_size: int
        context: _ResumeUploadV1Context
        kwargs:
            key, file_name, params, metadata

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """
        key = kwargs.get('key', None)
        file_name = kwargs.get('file_name', None)
        params = kwargs.get('params', None)
        metadata = kwargs.get('metadata', None)
        mime_type = kwargs.get('mime_type', None)

        # sort contexts
        sorted_parts = sorted(context.parts, key=lambda part: part.part_no)
        body = ','.join((part.ctx for part in sorted_parts))

        ret, resp = None, None
        for up_host in context.up_hosts:
            url = self.__get_mkfile_url(
                up_host=up_host,
                data_size=data_size,
                mime_type=mime_type,
                key=key,
                file_name=file_name,
                params=params,
                metadata=metadata
            )
            ret, resp = qn_http_client.post(
                url=url,
                data=body,
                files=None,
                headers={
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if resp.ok() or not resp.need_retry():
                break
        self._try_delete_record(
            file_name,
            key,
            context,
            resp
        )
        return ret, resp

    def __upload_with_retrier(
        self,
        access_key,
        bucket_name,
        **upload_opts
    ):
        file_name = upload_opts.get('file_name', None)
        key = upload_opts.get('key', None)
        modify_time = upload_opts.get('modify_time', None)

        context = self._initial_context(
            key=key,
            file_name=file_name,
            modify_time=modify_time
        )
        preferred_endpoints = None
        if context.up_hosts:
            preferred_endpoints = [
                Endpoint.from_host(h)
                for h in context.up_hosts
            ]

        progress_record = None
        if all([
            self.upload_progress_recorder,
            file_name,
            key
        ]):
            progress_record = ProgressRecord(
                upload_api_version='v1',
                exists=functools.partial(
                    self.upload_progress_recorder.has_upload_record,
                    file_name=file_name,
                    key=key
                ),
                delete=functools.partial(
                    self.upload_progress_recorder.delete_upload_record,
                    file_name=file_name,
                    key=key
                )
            )

        retrier = get_default_retrier(
            regions_provider=self._get_regions_provider(
                access_key=access_key,
                bucket_name=bucket_name
            ),
            preferred_endpoints_provider=preferred_endpoints,
            progress_record=progress_record,
            accelerate_uploading=self.accelerate_uploading,
        )

        data = upload_opts.get('data')
        attempt = None
        for attempt in retrier:
            with attempt:
                upload_opts['up_endpoint'] = attempt.context.get('endpoint')
                attempt.result = self.__upload(
                    **upload_opts
                )
                ret, resp = attempt.result
                if resp.ok() and ret:
                    return attempt.result
                if (
                    not is_seekable(data) or
                    not resp.need_retry()
                ):
                    return attempt.result
                data.seek(0)

        if attempt is None:
            raise RuntimeError('Retrier is not working. attempt is None')

        return attempt.result

    def __upload(
        self,
        up_token,
        key,
        file_path,
        file_name,
        data,
        data_size,
        modify_time,
        mime_type,
        custom_vars,
        metadata,
        up_endpoint
    ):
        # initial_parts
        context, resp = self.initial_parts(
            up_token,
            key,
            file_path=file_path,
            file_name=file_name,
            data=data,
            data_size=data_size,
            modify_time=modify_time,
            up_endpoint=up_endpoint
        )

        # upload_parts
        try:
            if file_path:
                data_size = path.getsize(file_path)
                data = open(file_path, 'rb')
            elif isinstance(data, bytes):
                data_size = len(data)
                data = BytesIO(data)
            elif isinstance(data, str):
                data_size = len(data)
                data = BytesIO(b(data))
            ret, resp = self.upload_parts(
                up_token=up_token,
                context=context,
                data=data,
                data_size=data_size,

                key=key,
                file_name=file_name
            )
        finally:
            if file_path:
                data.close()

        if resp and not resp.ok():
            return ret, resp

        # complete_parts
        ret, resp = self.complete_parts(
            up_token=up_token,
            data_size=data_size,
            context=context,

            key=key,
            mime_type=mime_type,
            file_name=file_name,
            params=custom_vars,
            metadata=metadata
        )

        return ret, resp

    def upload(
        self,
        key,
        file_path=None,
        data=None,
        data_size=None,
        modify_time=None,

        part_size=None,
        mime_type=None,
        metadata=None,
        file_name=None,
        custom_vars=None,
        **kwargs
    ):
        """

        Parameters
        ----------
        key
        file_path
        data
        data_size
        modify_time

        part_size
        mime_type
        metadata
        file_name
        custom_vars

        kwargs:
            up_token: str
            crc32_int: int
            bucket_name: str
                is required if upload to another bucket
            expired: int
                option for generate up_token if not provide up_token. seconds
            policy: dict
                option for generate up_token if not provide up_token. details see `auth.Auth`
            strict_policy: bool
                option for generate up_token if not provide up_token

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """
        # part_size
        if part_size:
            logging.warning('ResumeUploader not support part_size. It is fixed to 4MB.')

        # up_token
        up_token = kwargs.get('up_token', None)
        if not up_token:
            kwargs.setdefault('up_token', self.get_up_token(**kwargs))
            access_key = self.auth.get_access_key()
        else:
            access_key, _, _ = Auth.up_token_decode(up_token)

        # bucket_name
        kwargs['bucket_name'] = Auth.get_bucket_name(up_token)

        # file_name
        if not file_name and file_path:
            file_name = path.basename(file_path)

        # upload
        return self.__upload_with_retrier(
            access_key=access_key,
            key=key,
            file_path=file_path,
            data=data,
            data_size=data_size,
            modify_time=modify_time,
            mime_type=mime_type,
            metadata=metadata,
            file_name=file_name,
            custom_vars=custom_vars,
            **kwargs
        )

    def __upload_part(
        self,
        data,
        chunk_info,
        up_hosts,
        up_token,
        lock
    ):
        """
        Parameters
        ----------
        data: IOBase
        chunk_info: ChunkInfo
        up_hosts: list[str]
        up_token: str
        lock: Lock

        Returns
        -------
        part: _ResumeUploadV2Part
        resp: ResponseInfo
        """
        if not up_hosts:
            raise ValueError('Must provide one up host at least')

        chunked_data = IOChunked(
            base_io=data,
            chunk_offset=chunk_info.chunk_offset,
            chunk_size=chunk_info.chunk_size,
            lock=lock
        )
        chunk_crc32 = io_crc32(chunked_data)
        chunked_data.seek(0)
        part, resp = None, None
        for up_host in up_hosts:
            url = '/'.join([
                up_host,
                'mkblk', str(chunk_info.chunk_size)
            ])
            ret, resp = qn_http_client.post(
                url=url,
                data=chunked_data,
                files=None,
                headers={
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if resp.ok() and ret:
                if ret.get('crc32', 0) != chunk_crc32:
                    return None, resp
                part = _ResumeUploadV1Part(
                    part_no=chunk_info.chunk_no,
                    ctx=ret.get('ctx', ''),
                    expired_at=ret.get('expired_at', 0),
                )
                return part, resp
            if (
                not is_seekable(chunked_data) or
                not resp.need_retry()
            ):
                return part, resp
            chunked_data.seek(0)
        return part, resp

    def __get_mkfile_url(
        self,
        up_host,
        data_size,
        mime_type=None,
        key=None,
        file_name=None,
        params=None,
        metadata=None
    ):
        """
        Parameters
        ----------
        up_host: str
        data_size: int
        mime_type: str or None
        key: str or None
        file_name: str or None
        params: dict or None
        metadata: dict or None

        Returns
        -------
        str
        """
        url_base = [up_host, 'mkfile', str(data_size)]
        url_params = []

        if mime_type:
            url_params.append(('mimeType', mime_type))

        if key:
            url_params.append(('key', key))

        if file_name:
            url_params.append(('fname', file_name))

        if params:
            url_params.extend(params.items())

        if metadata:
            url_params.extend(
                (k, v)
                for k, v in metadata.items()
                if k.startswith('x-qn-meta-')
            )

        url_params_iter = chain.from_iterable(
            (str(k), urlsafe_base64_encode(str(v)))
            for k, v in url_params
        )

        return '/'.join(
            chain(
                url_base,
                url_params_iter
            )
        )


# use dataclass instead namedtuple if min version of python update to 3.7
_ResumeUploadV1Part = namedtuple(
    'ResumeUploadV1Part',
    [
        'part_no',
        'ctx',
        'expired_at',
    ]
)

_ResumeUploadV1Context = namedtuple(
    'ResumeUploadV1Context',
    [
        'up_hosts',
        'part_size',
        'parts',
        'modify_time',  # the file last modify time
        'resumed'
    ]
)



import os
import io
from collections import namedtuple

from qiniu.compat import is_seekable


ChunkInfo = namedtuple(
    'ChunkInfo',
    [
        'chunk_no',
        'chunk_offset',
        'chunk_size'
    ]
)


class IOChunked(io.IOBase):
    def __init__(
        self,
        base_io,
        chunk_offset,
        chunk_size,
        lock,
        buffer_size=4 * (1024 ** 2)  # 4MB just for demo
    ):
        if not is_seekable(base_io):
            raise TypeError('"base_io" must be seekable')
        self.__base_io = base_io
        self.__chunk_start = chunk_offset
        self.__chunk_size = chunk_size
        self.__chunk_end = chunk_offset + chunk_size
        self.__lock = lock
        self.__chunk_pos = 0

        self.buffer_size = min(buffer_size, chunk_size)

    def readable(self):
        return self.__base_io.readable()

    def seekable(self):
        return True

    def seek(self, offset, whence=0):
        if not self.seekable():
            raise io.UnsupportedOperation('does not support seek')
        if whence == os.SEEK_SET:
            if offset < 0:
                raise ValueError('offset should be zero or positive if whence is 0')
            self.__chunk_pos = offset
        elif whence == os.SEEK_CUR:
            self.__chunk_pos += offset
        elif whence == os.SEEK_END:
            if offset > 0:
                raise ValueError('offset should be zero or negative if whence is 2')
            self.__chunk_pos = self.__chunk_size + offset
        else:
            raise ValueError('whence should be 0, 1 or 2')
        self.__chunk_pos = max(
            0,
            min(self.__chunk_size, self.__chunk_pos)
        )

    def tell(self):
        return self.__chunk_pos

    def read(self, size):
        if self.__curr_base_pos >= self.__chunk_end:
            return b''
        read_size = max(self.buffer_size, size)
        read_size = min(self.__rest_chunk_size, read_size)

        # -- ignore size argument --
        with self.__lock:
            self.__base_io.seek(self.__curr_base_pos)
            data = self.__base_io.read(read_size)

        self.__chunk_pos += len(data)
        return data

    def __len__(self):
        return self.__chunk_size

    @property
    def __curr_base_pos(self):
        return self.__chunk_start + self.__chunk_pos

    @property
    def __rest_chunk_size(self):
        return self.__chunk_end - self.__curr_base_pos



import platform as _platform

from .base import Middleware


class UserAgentMiddleware(Middleware):
    def __init__(self, sdk_version):
        sys_info = '{0}; {1}'.format(_platform.system(), _platform.machine())
        python_ver = _platform.python_version()

        user_agent = 'QiniuPython/{0} ({1}; ) Python/{2}'.format(
            sdk_version, sys_info, python_ver)

        self.user_agent = user_agent

    def __call__(self, request, nxt):
        if not request.headers:
            request.headers = {}
        request.headers['User-Agent'] = self.user_agent
        return nxt(request)



# -*- coding: utf-8 -*-
import hashlib
import os
import time

from qiniu import config, http
from qiniu.auth import Auth
from qiniu.compat import json
from qiniu.utils import _file_iter, crc32, rfc_from_timestamp, urlsafe_base64_encode, deprecated

from qiniu.services.storage.upload_progress_recorder import UploadProgressRecorder


@deprecated("use uploader instead")
class _Resume(object):
    """deprecated 断点续上传类

    该类主要实现了分块上传，断点续上，以及相应地创建块和创建文件过程，详细规格参考：
    https://developer.qiniu.com/kodo/api/mkblk
    https://developer.qiniu.com/kodo/api/mkfile

    Attributes:
        up_token:                   上传凭证
        key:                        上传文件名
        input_stream:               上传二进制流
        data_size:                  上传流大小
        params:                     自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
        mime_type:                  上传数据的mimeType
        progress_handler:           上传进度
        upload_progress_recorder:   记录上传进度，用于断点续传
        modify_time:                上传文件修改日期
        hostscache_dir：            host请求 缓存文件保存位置
        version                     分片上传版本 目前支持v1/v2版本 默认v1
        part_size                   分片上传v2必传字段 分片大小范围为1 MB - 1 GB
        bucket_name                 分片上传v2字段 空间名称
    """

    def __init__(self, up_token, key, input_stream, file_name, data_size, hostscache_dir, params, mime_type,
                 progress_handler, upload_progress_recorder, modify_time, keep_last_modified, part_size=None,
                 version=None, bucket_name=None, metadata=None):
        """初始化断点续上传"""
        self.up_token = up_token
        self.key = key
        self.input_stream = input_stream
        self.file_name = file_name
        self.size = data_size
        self.hostscache_dir = hostscache_dir
        self.blockStatus = []
        self.params = params
        self.mime_type = mime_type
        self.progress_handler = progress_handler
        self.upload_progress_recorder = upload_progress_recorder or UploadProgressRecorder()
        self.modify_time = modify_time or time.time()
        self.keep_last_modified = keep_last_modified
        self.version = version or 'v1'
        self.part_size = part_size or config._BLOCK_SIZE
        self.bucket_name = bucket_name
        self.metadata = metadata

    def record_upload_progress(self, offset):
        record_data = {
            'size': self.size,
            'offset': offset,
        }
        if self.version == 'v1':
            record_data['contexts'] = [
                {
                    'ctx': block['ctx'],
                    'expired_at': block['expired_at'] if 'expired_at' in block else 0
                } for block in self.blockStatus
            ]
        elif self.version == 'v2':
            record_data['etags'] = self.blockStatus
            record_data['expired_at'] = self.expiredAt
            record_data['upload_id'] = self.uploadId
        if self.modify_time:
            record_data['modify_time'] = self.modify_time
        self.upload_progress_recorder.set_upload_record(self.file_name, self.key, record_data)

    def recovery_from_record(self):
        record = self.upload_progress_recorder.get_upload_record(self.file_name, self.key)
        if not record:
            if self.version == 'v1':
                return 0
            elif self.version == 'v2':
                return 0, None, None
        try:
            if not record['modify_time'] or record['size'] != self.size or \
                    record['modify_time'] != self.modify_time:
                if self.version == 'v1':
                    return 0
                elif self.version == 'v2':
                    return 0, None, None
        except KeyError:
            if self.version == 'v1':
                return 0
            elif self.version == 'v2':
                return 0, None, None
        if self.version == 'v1':
            if not record.__contains__('contexts') or len(record['contexts']) == 0:
                return 0
            self.blockStatus = [
                # 兼容旧版本的 ctx 持久化 ≤v7.10.0
                ctx if type(ctx) is dict else {'ctx': ctx, 'expired_at': 0}
                for ctx in record['contexts']
            ]
            return record['offset']
        elif self.version == 'v2':
            if not record.__contains__('etags') or len(record['etags']) == 0 or \
                    not record.__contains__('expired_at') or float(record['expired_at']) < time.time() or \
                    not record.__contains__('upload_id'):
                return 0, None, None
            self.blockStatus = record['etags']
            return record['offset'], record['upload_id'], record['expired_at']

    def upload(self):
        """上传操作"""
        if self.version == 'v1':
            return self._upload_v1()
        elif self.version == 'v2':
            return self._upload_v2()
        else:
            raise ValueError("version must choose v1 or v2 !")

    def _upload_v1(self):
        self.blockStatus = []
        self.recovery_index = 1
        self.expiredAt = None
        self.uploadId = None
        self.get_bucket()
        self.part_size = config._BLOCK_SIZE

        host = self.get_up_host()
        offset = self.recovery_from_record()
        is_resumed = offset > 0

        # 检查原来的分片是否过期，如有则重传该分片
        for index, block_status in enumerate(self.blockStatus):
            if block_status.get('expired_at', 0) > time.time():
                self.input_stream.seek(self.part_size, os.SEEK_CUR)
            else:
                block = self.input_stream.read(self.part_size)
                response, ok = self._make_block_with_retry(block, host)
                ret, info = response
                if not ok:
                    return ret, info
                self.blockStatus[index] = ret
            self.record_upload_progress(offset)

        # 从断点位置上传
        for block in _file_iter(self.input_stream, self.part_size, offset):
            length = len(block)
            response, ok = self._make_block_with_retry(block, host)
            ret, info = response
            if not ok:
                return ret, info

            self.blockStatus.append(ret)
            offset += length
            self.record_upload_progress(offset)
            if callable(self.progress_handler):
                self.progress_handler(((len(self.blockStatus) - 1) * self.part_size) + len(block), self.size)

        ret, info = self.make_file(host)
        if info.status_code == 200 or info.status_code == 701:
            self.upload_progress_recorder.delete_upload_record(self.file_name, self.key)
        if info.status_code == 701 and is_resumed:
            return self.upload()
        return ret, info

    def _upload_v2(self):
        self.blockStatus = []
        self.recovery_index = 1
        self.expiredAt = None
        self.uploadId = None
        self.get_bucket()
        host = self.get_up_host()

        offset, self.uploadId, self.expiredAt = self.recovery_from_record()
        is_resumed = False
        if offset > 0 and self.blockStatus != [] and self.uploadId is not None \
                and self.expiredAt is not None:
            self.recovery_index = self.blockStatus[-1]['partNumber'] + 1
            is_resumed = True
        else:
            self.recovery_index = 1
            init_url = self.block_url_v2(host, self.bucket_name)
            self.uploadId, self.expiredAt = self.init_upload_task(init_url)

        for index, block in enumerate(_file_iter(self.input_stream, self.part_size, offset)):
            length = len(block)
            index_ = index + self.recovery_index
            url = self.block_url_v2(host, self.bucket_name) + '/%s/%d' % (self.uploadId, index_)
            ret, info = self.make_block_v2(block, url)
            if info.status_code == 612:
                self.upload_progress_recorder.delete_upload_record(self.file_name, self.key)
            if info.status_code == 612 and is_resumed:
                return self.upload()
            if ret is None and not info.need_retry():
                return ret, info
            if info.connect_failed():
                if config.get_default('default_zone').up_host_backup:
                    host = config.get_default('default_zone').up_host_backup
                else:
                    host = config.get_default('default_zone') \
                        .get_up_host_backup_by_token(self.up_token, self.hostscache_dir)

            if info.need_retry():
                url = self.block_url_v2(host, self.bucket_name) + '/%s/%d' % (self.uploadId, index + 1)
                ret, info = self.make_block_v2(block, url)
                if info.status_code == 612:
                    self.upload_progress_recorder.delete_upload_record(self.file_name, self.key)
                if info.status_code == 612 and is_resumed:
                    return self.upload()
                if ret is None:
                    return ret, info
            del ret['md5']
            ret['partNumber'] = index_
            self.blockStatus.append(ret)
            offset += length
            self.record_upload_progress(offset)
            if callable(self.progress_handler):
                self.progress_handler(((len(self.blockStatus) - 1) * self.part_size) + len(block), self.size)

        make_file_url = self.block_url_v2(host, self.bucket_name) + '/%s' % self.uploadId
        ret, info = self.make_file_v2(
            self.blockStatus,
            make_file_url,
            self.file_name,
            self.mime_type,
            self.params,
            self.metadata)
        if info.status_code == 200 or info.status_code == 612:
            self.upload_progress_recorder.delete_upload_record(self.file_name, self.key)
        if info.status_code == 612 and is_resumed:
            return self.upload()
        return ret, info

    def make_file_v2(self, block_status, url, file_name=None, mime_type=None, customVars=None, metadata=None):
        """completeMultipartUpload"""
        parts = self.get_parts(block_status)
        headers = {
            'Content-Type': 'application/json',
        }
        data = {
            'parts': parts,
            'fname': file_name,
            'mimeType': mime_type,
            'customVars': customVars,
            'metadata': metadata
        }
        return self.post_with_headers(url, json.dumps(data), headers=headers)

    def get_up_host(self):
        if config.get_default('default_zone').up_host:
            host = config.get_default('default_zone').up_host
        else:
            host = config.get_default('default_zone').get_up_host_by_token(self.up_token, self.hostscache_dir)
        return host

    def _make_block_with_retry(self, block_data, up_host):
        length = len(block_data)
        crc = crc32(block_data)
        ret, info = self.make_block(block_data, length, up_host)
        if ret is None and not info.need_retry():
            return (ret, info), False
        if info.connect_failed():
            if config.get_default('default_zone').up_host_backup:
                up_host = config.get_default('default_zone').up_host_backup
            else:
                up_host = config.get_default('default_zone') \
                    .get_up_host_backup_by_token(self.up_token, self.hostscache_dir)
            if info.need_retry() or crc != ret['crc32']:
                ret, info = self.make_block(block_data, length, up_host)
                if ret is None or crc != ret['crc32']:
                    return (ret, info), False
        return (ret, info), True

    def make_block(self, block, block_size, host):
        """创建块"""
        url = self.block_url(host, block_size)
        return self.post(url, block)

    def make_block_v2(self, block, url):
        headers = {
            'Content-Type': 'application/octet-stream',
            'Content-MD5': hashlib.md5(block).hexdigest(),
        }
        return self.put(url, block, headers)

    def block_url(self, host, size):
        return '{0}/mkblk/{1}'.format(host, size)

    def block_url_v2(self, host, bucket_name):
        encoded_object_name = urlsafe_base64_encode(self.key) if self.key is not None else '~'
        return '{0}/buckets/{1}/objects/{2}/uploads'.format(host, bucket_name, encoded_object_name)

    def file_url(self, host):
        url = ['{0}/mkfile/{1}'.format(host, self.size)]
        if self.mime_type:
            url.append('mimeType/{0}'.format(urlsafe_base64_encode(self.mime_type)))

        if self.key is not None:
            url.append('key/{0}'.format(urlsafe_base64_encode(self.key)))

        if self.file_name is not None:
            url.append('fname/{0}'.format(urlsafe_base64_encode(self.file_name)))

        if self.params:
            for k, v in self.params.items():
                url.append('{0}/{1}'.format(k, urlsafe_base64_encode(v)))

        if self.modify_time and self.keep_last_modified:
            url.append(
                "x-qn-meta-!Last-Modified/{0}".format(urlsafe_base64_encode(rfc_from_timestamp(self.modify_time))))

        if self.metadata:
            for k, v in self.metadata.items():
                if k.startswith('x-qn-meta-'):
                    url.append(
                        "{0}/{1}".format(k, urlsafe_base64_encode(v)))

        url = '/'.join(url)
        return url

    def make_file(self, host):
        """创建文件"""
        url = self.file_url(host)
        body = ','.join([status['ctx'] for status in self.blockStatus])
        return self.post(url, body)

    def init_upload_task(self, url):
        body, resp = self.post(url, '')
        if body is not None:
            return body['uploadId'], body['expireAt']
        else:
            return None, None

    def post(self, url, data):
        return http._post_with_token(url, data, self.up_token)

    def post_with_headers(self, url, data, headers):
        return http._post_with_token_and_headers(url, data, self.up_token, headers)

    def put(self, url, data, headers):
        return http._put_with_token_and_headers(url, data, self.up_token, headers)

    def get_parts(self, block_status):
        return sorted(block_status, key=lambda i: i['partNumber'])

    def get_bucket(self):
        if not self.bucket_name:
            bucket_name = Auth.get_bucket_name(self.up_token)
            if bucket_name:
                self.bucket_name = bucket_name



# -*- coding: utf-8 -*-
import os
import time

import requests

from qiniu.auth import QiniuMacAuth, QiniuMacRequestsAuth
from qiniu.compat import basestring, urlencode

from .constants import DEFAULT_TEMPLATE
from .errors import SandboxError, TemplateBuildError
from .resources import KodoResource
from .util import (
    encode_path,
    json_dumps,
    normalize_endpoint,
    parse_json_response,
)


_UNSET = object()


try:
    from time import monotonic as _monotonic_time
except ImportError:
    _monotonic_time = time.time


def _to_dict(value):
    if hasattr(value, 'to_dict'):
        return value.to_dict()
    return value


def _normalize_injection(injection):
    if hasattr(injection, 'to_dict'):
        injection = injection.to_dict()
    if not isinstance(injection, dict):
        return injection
    data = dict(injection)
    if 'apiKey' in data and 'api_key' not in data:
        data['api_key'] = data.pop('apiKey')
    if 'baseUrl' in data and 'base_url' not in data:
        data['base_url'] = data.pop('baseUrl')
    if 'ifHeaders' in data and 'if_headers' not in data:
        data['if_headers'] = data.pop('ifHeaders')
    if 'ifQueries' in data and 'if_queries' not in data:
        data['if_queries'] = data.pop('ifQueries')
    if 'ruleId' in data and 'ruleID' not in data:
        data['ruleID'] = data.pop('ruleId')
    if 'rule_id' in data and 'ruleID' not in data:
        data['ruleID'] = data.pop('rule_id')
    return data


def _normalize_resources(resources):
    if resources is None:
        return None
    return [_to_dict(resource) for resource in resources]


def _has_kodo_resource(resources):
    for resource in resources or []:
        if isinstance(resource, KodoResource):
            return True
        data = _to_dict(resource)
        if isinstance(data, dict) and data.get('type') == 'kodo':
            return True
    return False


def _has_saved_injection_rule(injections):
    for injection in injections or []:
        data = _normalize_injection(injection)
        if not isinstance(data, dict):
            continue
        if data.get('type') == 'id':
            return True
        if data.get('ruleID') or data.get('rule_id') or data.get('id'):
            return True
    return False


def _normalize_sandbox_create_options(template=None, **opts):
    body = {'templateID': (
        opts.pop('templateID', None) or
        opts.pop('template_id', None) or
        opts.pop('template', None) or
        template or
        DEFAULT_TEMPLATE
    )}
    for key in (
        'timeout',
        'autoPause',
        'secure',
        'network',
        'metadata',
        'mcp',
    ):
        if opts.get(key) is not None:
            body[key] = opts.get(key)
    if opts.get('auto_pause') is not None:
        body['autoPause'] = opts.get('auto_pause')
    allow_internet_access = _single_alias_value(
        opts, 'allow_internet_access', 'allowInternetAccess')
    if allow_internet_access is not None:
        body['allowInternetAccess'] = allow_internet_access
    envs = _single_alias_value(opts, 'envs', 'envVars')
    if envs is not None:
        body['envVars'] = envs
    if opts.get('lifecycle') is not None:
        lifecycle = opts.get('lifecycle') or {}
        body['lifecycle'] = lifecycle
        on_timeout = lifecycle.get('on_timeout') or lifecycle.get('onTimeout')
        if on_timeout == 'pause':
            body['autoPause'] = True
    if opts.get('injections') is not None:
        body['injections'] = [_normalize_injection(
            item) for item in opts.get('injections')]
    if opts.get('resources') is not None:
        body['resources'] = _normalize_resources(opts.get('resources'))
    return body


def _single_alias_value(opts, *names):
    present = [name for name in names if opts.get(name) is not None]
    if len(present) > 1:
        raise SandboxError(
            'Conflicting sandbox create options: {0}'.format(
                ', '.join(present)))
    if present:
        return opts.get(present[0])
    return None


def _require_sandbox_id(sandbox_id):
    if not sandbox_id:
        raise SandboxError('sandbox_id is required')
    return sandbox_id


def _normalize_list_options(opts):
    opts = dict(opts or {})
    query = opts.pop('query', None) or {}
    metadata = query.get('metadata')
    if metadata is not None:
        opts['metadata'] = metadata
    if isinstance(opts.get('metadata'), dict):
        opts['metadata'] = urlencode(opts.get('metadata'))
    if query.get('state') is not None:
        opts['state'] = query.get('state')
    return opts


def _sandbox_api_key_from_env():
    return (
        os.getenv('QINIU_SANDBOX_API_KEY') or
        os.getenv('QINIU_API_KEY') or
        os.getenv('E2B_API_KEY')
    )


class SandboxClient(object):
    def __init__(self, endpoint=None, api_url=None, api_key=None,
                 access_token=None, mac=None, access_key=None,
                 secret_key=None, session=None, timeout=None, **opts):
        access_key = access_key or os.getenv('QINIU_SANDBOX_ACCESS_KEY')
        secret_key = secret_key or os.getenv('QINIU_SANDBOX_SECRET_KEY')
        if (access_key and not secret_key) or (secret_key and not access_key):
            raise SandboxError(
                'Both access_key and secret_key must be provided')
        self.endpoint = normalize_endpoint(endpoint or api_url)
        self.api_key = api_key or _sandbox_api_key_from_env()
        self.access_token = access_token or os.getenv(
            'QINIU_SANDBOX_ACCESS_TOKEN')
        self.mac = mac
        if self.mac is None and access_key and secret_key:
            self.mac = QiniuMacAuth(access_key, secret_key)
        self.session = session or requests.Session()
        self.timeout = timeout if timeout is not None else 30

    def _headers(self, auth_type=None):
        headers = {'Content-Type': 'application/json'}
        if self.api_key:
            headers['X-API-Key'] = self.api_key
        if auth_type == 'qiniu':
            return headers
        if auth_type == 'accessToken':
            if not self.access_token:
                raise SandboxError(
                    'access_token is required for this operation')
            headers['Authorization'] = 'Bearer {0}'.format(
                self.access_token)
            return headers
        if self.api_key:
            headers['Authorization'] = 'Bearer {0}'.format(self.api_key)
        elif self.access_token:
            headers['Authorization'] = 'Bearer {0}'.format(self.access_token)
        return headers

    def _auth(self, auth_type=None):
        if auth_type == 'qiniu' or (
                not self.api_key and not self.access_token and self.mac):
            if not self.mac:
                raise SandboxError(
                    'Qiniu AK/SK credentials are required for this operation'
                )
            return QiniuMacRequestsAuth(self.mac)
        return None

    def _request(self, method, path, params=None, body=_UNSET,
                 auth_type=None, empty=False):
        url = self.endpoint + path
        data = None if body is _UNSET else json_dumps(body)
        headers = self._headers(auth_type)
        auth = self._auth(auth_type)
        request = requests.Request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
        )
        prepared = self.session.prepare_request(request)
        try:
            response = self.session.send(prepared, timeout=self.timeout)
        except requests.RequestException as err:
            raise SandboxError('Sandbox API request failed: {0}'.format(err))
        if response.status_code < 200 or response.status_code >= 300:
            response_data = None
            try:
                response_data = response.json()
            except ValueError:
                try:
                    response_data = response.text
                except Exception:
                    response_data = getattr(response, 'content', None)
            message = 'Sandbox API request failed with status {0}'.format(
                response.status_code
            )
            if isinstance(response_data, dict):
                err_msg = response_data.get(
                    'message') or response_data.get('error')
                if isinstance(err_msg, dict):
                    err_msg = err_msg.get('message') or err_msg.get('error')
                if err_msg:
                    message += ': {0}'.format(err_msg)
            elif isinstance(response_data, basestring) and response_data:
                if len(response_data) > 200:
                    response_data = response_data[:200] + '...'
                message += ': {0}'.format(response_data)
            raise SandboxError(message, response, response_data)
        if empty:
            return None
        return parse_json_response(response)

    def list_sandboxes(self, **opts):
        return self._request('GET', '/sandboxes', params=opts)

    listSandboxes = list_sandboxes

    def list_sandboxes_v2(self, **opts):
        return self._request(
            'GET',
            '/v2/sandboxes',
            params=_normalize_list_options(opts))

    listSandboxesV2 = list_sandboxes_v2
    list = list_sandboxes_v2

    def create_sandbox(self, template=None, **opts):
        body = _normalize_sandbox_create_options(template, **opts)
        auth_type = 'qiniu' if (
            _has_kodo_resource(body.get('resources')) or
            _has_saved_injection_rule(body.get('injections'))
        ) else None
        return self._request(
            'POST',
            '/sandboxes',
            body=body,
            auth_type=auth_type)

    createSandbox = create_sandbox
    create = create_sandbox

    def get_sandbox(self, sandbox_id):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'GET', '/sandboxes/{0}'.format(encode_path(sandbox_id)))

    getSandbox = get_sandbox
    get_info = get_sandbox
    getInfo = get_sandbox

    def delete_sandbox(self, sandbox_id):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'DELETE',
            '/sandboxes/{0}'.format(encode_path(sandbox_id)),
            empty=True,
        )

    deleteSandbox = delete_sandbox
    kill_sandbox = delete_sandbox
    killSandbox = delete_sandbox
    kill = delete_sandbox

    def pause_sandbox(self, sandbox_id):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'POST',
            '/sandboxes/{0}/pause'.format(encode_path(sandbox_id)),
            body={},
            empty=True,
        )

    pauseSandbox = pause_sandbox

    def resume_sandbox(self, sandbox_id, **opts):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'POST',
            '/sandboxes/{0}/resume'.format(encode_path(sandbox_id)),
            body=opts,
        )

    resumeSandbox = resume_sandbox

    def connect_sandbox(self, sandbox_id, timeout=15):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'POST',
            '/sandboxes/{0}/connect'.format(encode_path(sandbox_id)),
            body={'timeout': timeout},
        )

    connectSandbox = connect_sandbox
    connect = connect_sandbox

    def update_sandbox_timeout(self, sandbox_id, timeout=None, **opts):
        _require_sandbox_id(sandbox_id)
        if timeout is None:
            timeout = opts.get('timeout')
        return self._request(
            'POST',
            '/sandboxes/{0}/timeout'.format(encode_path(sandbox_id)),
            body={'timeout': timeout},
            empty=True,
        )

    updateSandboxTimeout = update_sandbox_timeout
    set_timeout = update_sandbox_timeout
    setTimeout = update_sandbox_timeout

    def refresh_sandbox(self, sandbox_id, **opts):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'POST',
            '/sandboxes/{0}/refreshes'.format(encode_path(sandbox_id)),
            body=opts,
            empty=True,
        )

    refreshSandbox = refresh_sandbox

    def update_sandbox(self, sandbox_id, **opts):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'PATCH',
            '/sandboxes/{0}'.format(encode_path(sandbox_id)),
            body=opts,
        )

    updateSandbox = update_sandbox

    def get_sandbox_metrics(self, sandbox_id, **opts):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'GET',
            '/sandboxes/{0}/metrics'.format(encode_path(sandbox_id)),
            params=opts,
        )

    getSandboxMetrics = get_sandbox_metrics
    get_metrics = get_sandbox_metrics
    getMetrics = get_sandbox_metrics

    def get_sandboxes_metrics(self, sandbox_ids):
        if not sandbox_ids:
            raise SandboxError('At least one sandbox ID must be provided')
        values = sandbox_ids
        if isinstance(sandbox_ids, dict):
            values = sandbox_ids.get(
                'sandbox_ids') or sandbox_ids.get('sandboxIDs')
            if values is None:
                values = [sandbox_ids]
        if values is None:
            raise SandboxError('At least one sandbox ID must be provided')
        if isinstance(values, (list, tuple, set)):
            values = list(values)
        elif hasattr(values, '__iter__') and not isinstance(
                values, (basestring, dict)):
            values = list(values)
        else:
            values = [values]
        ids = []
        for value in values:
            if isinstance(value, basestring):
                ids.append(value)
            elif isinstance(value, dict):
                ids.append(value.get('sandboxId') or value.get(
                    'sandboxID') or value.get('id'))
            elif hasattr(value, 'sandbox_id'):
                ids.append(value.sandbox_id)
        ids = [item for item in ids if item]
        if not ids:
            raise SandboxError('At least one sandbox ID must be provided')
        return self._request(
            'GET',
            '/sandboxes/metrics',
            params={
                'sandbox_ids': ','.join(ids)})

    getSandboxesMetrics = get_sandboxes_metrics

    def get_sandbox_logs(self, sandbox_id, **opts):
        _require_sandbox_id(sandbox_id)
        return self._request(
            'GET',
            '/sandboxes/{0}/logs'.format(encode_path(sandbox_id)),
            params=opts,
        )

    getSandboxLogs = get_sandbox_logs
    get_logs = get_sandbox_logs
    getLogs = get_sandbox_logs

    def create_template(self, **opts):
        return self._request('POST', '/v3/templates', body=opts)

    createTemplate = create_template
    createTemplateV3 = create_template

    def create_template_v2(self, **opts):
        return self._request('POST', '/v2/templates', body=opts)

    createTemplateV2 = create_template_v2

    def list_templates(self, **opts):
        return self._request('GET', '/templates', params=opts)

    listTemplates = list_templates

    def list_default_templates(self):
        return self._request('GET', '/default-templates')

    listDefaultTemplates = list_default_templates

    def get_template(self, template_id, **opts):
        return self._request(
            'GET',
            '/templates/{0}'.format(encode_path(template_id)),
            params=opts,
        )

    getTemplate = get_template

    def delete_template(self, template_id):
        if not template_id:
            raise SandboxError('template_id is required')
        return self._request(
            'DELETE',
            '/templates/{0}'.format(encode_path(template_id)),
            empty=True,
        )

    deleteTemplate = delete_template

    def update_template(self, template_id, **opts):
        return self._request(
            'PATCH',
            '/templates/{0}'.format(encode_path(template_id)),
            body=opts,
        )

    updateTemplate = update_template

    def rebuild_template(self, template_id, **opts):
        return self._request(
            'POST',
            '/templates/{0}'.format(encode_path(template_id)),
            body=opts,
            auth_type='accessToken',
        )

    rebuildTemplate = rebuild_template

    def get_template_build_status(self, template_id, build_id, **opts):
        return self._request(
            'GET',
            '/templates/{0}/builds/{1}/status'.format(
                encode_path(template_id),
                encode_path(build_id),
            ),
            params=opts,
        )

    getTemplateBuildStatus = get_template_build_status

    def get_template_build_logs(self, template_id, build_id, **opts):
        return self._request(
            'GET',
            '/templates/{0}/builds/{1}/logs'.format(
                encode_path(template_id),
                encode_path(build_id),
            ),
            params=opts,
        )

    getTemplateBuildLogs = get_template_build_logs

    def start_template_build(self, template_id, build_id, **opts):
        return self._request(
            'POST',
            '/v2/templates/{0}/builds/{1}'.format(
                encode_path(template_id),
                encode_path(build_id),
            ),
            body=opts,
            empty=True,
        )

    startTemplateBuild = start_template_build
    startTemplateBuildV2 = start_template_build

    def assign_template_tags(self, **opts):
        return self._request('POST', '/templates/tags', body=opts)

    assignTemplateTags = assign_template_tags

    def delete_template_tags(self, **opts):
        return self._request(
            'DELETE',
            '/templates/tags',
            body=opts,
            empty=True)

    deleteTemplateTags = delete_template_tags

    def get_template_by_alias(self, alias):
        return self._request(
            'GET', '/templates/aliases/{0}'.format(encode_path(alias)))

    getTemplateByAlias = get_template_by_alias

    def list_injection_rules(self):
        return self._request('GET', '/injection-rules', auth_type='qiniu')

    listInjectionRules = list_injection_rules

    def create_injection_rule(self, **opts):
        if opts.get('injection') is not None:
            opts['injection'] = _normalize_injection(opts.get('injection'))
        return self._request(
            'POST',
            '/injection-rules',
            body=opts,
            auth_type='qiniu',
        )

    createInjectionRule = create_injection_rule

    def get_injection_rule(self, rule_id):
        return self._request(
            'GET',
            '/injection-rules/{0}'.format(encode_path(rule_id)),
            auth_type='qiniu',
        )

    getInjectionRule = get_injection_rule

    def update_injection_rule(self, rule_id, **opts):
        if opts.get('injection') is not None:
            opts['injection'] = _normalize_injection(opts.get('injection'))
        return self._request(
            'PUT',
            '/injection-rules/{0}'.format(encode_path(rule_id)),
            body=opts,
            auth_type='qiniu',
        )

    updateInjectionRule = update_injection_rule

    def delete_injection_rule(self, rule_id):
        return self._request(
            'DELETE',
            '/injection-rules/{0}'.format(encode_path(rule_id)),
            auth_type='qiniu',
            empty=True,
        )

    deleteInjectionRule = delete_injection_rule

    def wait_for_build(self, template_id, build_id, interval=1, timeout=60):
        start = _monotonic_time()
        while True:
            try:
                info = self.get_template_build_status(template_id, build_id)
                if info and info.get('status') in ('ready', 'error'):
                    if info.get('status') == 'error':
                        message = (
                            (
                                isinstance(info.get('error'), dict) and
                                info.get('error').get('message')
                            ) or
                            info.get('error') or
                            info.get('message') or
                            'Sandbox template build failed'
                        )
                        raise TemplateBuildError(message, data=info)
                    return info
            except SandboxError as err:
                if isinstance(err, TemplateBuildError) or (
                        err.status_code is not None and
                        err.status_code >= 400 and err.status_code < 500):
                    raise
            if _monotonic_time() - start >= timeout:
                raise SandboxError('Sandbox template build polling timed out')
            time.sleep(interval)

    waitForBuild = wait_for_build
